#!./env/bin/python3
# -*- coding: utf-8 -*-
# Time-stamp: "2024-07-11 06:06:21 (ywatanabe)"
# /home/ywatanabe/proj/mngs_repo/src/mngs/plt/ax/_add_marginal_ax.py


"""
This script does XYZ.
"""


"""
Imports
"""
import os
import re
import sys

import matplotlib
import matplotlib.pyplot as plt
import seaborn as sns
import importlib

import mngs

importlib.reload(mngs)

import numpy as np
import pandas as pd
import torch
import torch.nn as nn
import torch.nn.functional as F
from icecream import ic
from natsort import natsorted
from glob import glob
from pprint import pprint
import warnings
import logging
from tqdm import tqdm
import xarray as xr
from mpl_toolkits.axes_grid1 import make_axes_locatable

# sys.path = ["."] + sys.path
# from scripts import utils, load

"""
Warnings
"""
# warnings.simplefilter("ignore", UserWarning)


"""
Config
"""
# CONFIG = mngs.gen.load_configs()


"""
Functions & Classes
"""


# def add_marginal_axes(ax):
#     divider = make_axes_locatable(ax)

#     ax_marg_x = divider.append_axes("top", size="20%", pad=0.1)
#     ax_marg_x.set_box_aspect(0.2)

#     ax_marg_y = divider.append_axes("right", size="20%", pad=0.1)
#     ax_marg_y.set_box_aspect(0.2 ** (-1))

#     return ax_marg_x, ax_marg_y


def add_marginal_ax(ax, place, size=0.2, pad=0.1):
    divider = make_axes_locatable(ax)

    size_perc_str = f"{size*100}%"
    if place in ["left", "right"]:
        size = 1.0 / size

    ax_marg = divider.append_axes(place, size=size_perc_str, pad=pad)
    ax_marg.set_box_aspect(size)

    return ax_marg


def main():
    pass


if __name__ == "__main__":
    # # Argument Parser
    # import argparse
    # parser = argparse.ArgumentParser(description='')
    # parser.add_argument('--var', '-v', type=int, default=1, help='')
    # parser.add_argument('--flag', '-f', action='store_true', default=False, help='')
    # args = parser.parse_args()

    # Main
    CONFIG, sys.stdout, sys.stderr, plt, CC = mngs.gen.start(
        sys, plt, verbose=False
    )
    main()
    mngs.gen.close(CONFIG, verbose=False, notify=False)

# EOF



#!./env/bin/python3
# -*- coding: utf-8 -*-
# Time-stamp: "2024-10-06 18:26:45 (ywatanabe)"
# /home/ywatanabe/proj/mngs/src/mngs/general/_dotdict.py

import json
from types import MappingProxyType
from typing import Any, Dict, Iterator, List, Union
from copy import deepcopy

# class DotDict:
#     def __init__(self, data=None):
#         self._data = {} if data is None else data
#         self._frozen = False

#     def __getattr__(self, key: str) -> Any:
#         try:
#             value = self._data[key]
#             if isinstance(value, dict) and not isinstance(value, DotDict):
#                 return DotDict(value)
#             return value
#         except KeyError:
#             raise AttributeError(f"'{self.__class__.__name__}' object has no attribute '{key}'")

#     def __setattr__(self, key: str, value: Any) -> None:
#         if key == '_data':
#             super().__setattr__(key, value)
#         else:
#             self._data[key] = value

#     def __getitem__(self, key: str) -> Any:
#         return self.__getattr__(key)

#     def __setitem__(self, key: str, value: Any) -> None:
#         self.__setattr__(key, value)

#     def freeze(self) -> None:
#         """Freeze the DotDict, making it immutable."""
#         self._frozen = True
#         for value in self._data.values():
#             if isinstance(value, DotDict):
#                 value.freeze()
#         self._data = MappingProxyType(self._data)

#     def unfreeze(self) -> None:
#         """Unfreeze the DotDict, making it mutable again."""
#         if self._frozen:
#             self._data = dict(self._data)
#             for value in self._data.values():
#                 if isinstance(value, DotDict):
#                     value.unfreeze()
#             self._frozen = False

#     def get(self, key: str, default: Any = None) -> Any:
#         return self._data.get(key, default)

#     def to_dict(self) -> Dict[str, Any]:
#         """Recursively converts DotDict and nested DotDict objects back to ordinary dictionaries."""
#         return {k: v.to_dict() if isinstance(v, DotDict) else v for k, v in self._data.items()}

#     def __str__(self) -> str:
#         """Returns a string representation of the DotDict."""
#         return json.dumps(self.to_dict(), indent=4)

#     def __repr__(self) -> str:
#         """Returns a string representation of the DotDict for debugging and development."""
#         return self.__str__()

#     def __len__(self) -> int:
#         """Returns the number of key-value pairs in the dictionary."""
#         return len(self._data)

#     def __deepcopy__(self, memo):
#         return DotDict(deepcopy(self._data, memo))

#     def keys(self) -> List[str]:
#         """Returns a list of all the keys in the dictionary."""
#         return list(self._data.keys())

#     def values(self) -> List[Any]:
#         """Returns a list of all the values in the dictionary."""
#         return list(self._data.values())

#     def items(self) -> List[tuple]:
#         """Returns a list of all the items (key, value pairs) in the dictionary."""
#         return list(self._data.items())

#     def update(self, dictionary: Dict[str, Any]) -> None:
#         """Updates the dictionary with the key-value pairs from another dictionary."""
#         if self._frozen:
#             raise AttributeError("Cannot modify frozen DotDict")
#         for key, value in dictionary.items():
#             if isinstance(value, dict):
#                 value = DotDict(value)
#             self._data[key] = value

#     def setdefault(self, key: str, default: Any = None) -> Any:
#         """Returns the value of the given key. If the key does not exist, insert the key with the specified default value."""
#         if self._frozen:
#             raise AttributeError("Cannot modify frozen DotDict")
#         return self._data.setdefault(key, default)

#     def pop(self, key: str, default: Any = None) -> Any:
#         """Removes the specified key and returns the corresponding value."""
#         if self._frozen:
#             raise AttributeError("Cannot modify frozen DotDict")
#         return self._data.pop(key, default)

#     def __contains__(self, key: str) -> bool:
#         """Checks if the DotDict contains the specified key."""
#         return key in self._data

#     def __iter__(self) -> Iterator[str]:
#         """Returns an iterator over the keys of the dictionary."""
#         return iter(self._data)

class DotDict:
    """
    A dictionary subclass that allows attribute-like access to keys.
    """

    def __init__(self, dictionary):
        for key, value in dictionary.items():
            if isinstance(value, dict):
                value = DotDict(value)
            setattr(self, key, value)

    def __getitem__(self, key):
        return getattr(self, key)

    def __setitem__(self, key, value):
        setattr(self, key, value)

    def get(self, key, default=None):
        return getattr(self, key, default)

    def to_dict(self):
        """
        Recursively converts DotDict and nested DotDict objects back to ordinary dictionaries.
        """
        result = {}
        for key, value in self.__dict__.items():
            if isinstance(value, DotDict):
                value = value.to_dict()
            result[key] = value
        return result

    def __str__(self):
        """
        Returns a string representation of the dotdict by converting it to a dictionary and pretty-printing it.
        """
        return json.dumps(self.to_dict(), indent=4)

    def __repr__(self):
        """
        Returns a string representation of the dotdict for debugging and development.
        """
        return self.__str__()

    def __len__(self):
        """
        Returns the number of key-value pairs in the dictionary.
        """
        return len(self.__dict__)

    def keys(self):
        """
        Returns a view object displaying a list of all the keys in the dictionary.
        """
        return self.__dict__.keys()

    def values(self):
        """
        Returns a view object displaying a list of all the values in the dictionary.
        """
        return self.__dict__.values()

    def items(self):
        """
        Returns a view object displaying a list of all the items (key, value pairs) in the dictionary.
        """
        return self.__dict__.items()

    def update(self, dictionary):
        """
        Updates the dictionary with the key-value pairs from another dictionary.
        """
        for key, value in dictionary.items():
            if isinstance(value, dict):
                value = DotDict(value)
            setattr(self, key, value)

    def setdefault(self, key, default=None):
        """
        Returns the value of the given key. If the key does not exist, insert the key with the specified default value.
        """
        if key not in self.__dict__:
            self[key] = default
        return self[key]

    def pop(self, key, default=None):
        """
        Removes the specified key and returns the corresponding value.
        """
        return self.__dict__.pop(key, default)

    def __contains__(self, key):
        """
        Checks if the dotdict contains the specified key.
        """
        return key in self.__dict__

    def __iter__(self):
        """
        Returns an iterator over the keys of the dictionary.
        """
        return iter(self.__dict__)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Time-stamp: "2024-09-28 11:02:59 (ywatanabe)"
# /home/ywatanabe/proj/_mngs_repo_openhands/src/mngs/plt/ax/_conf_mat.py

import sys
import matplotlib
import matplotlib.pyplot as plt
import seaborn as sns
import numpy as np
import pandas as pd
from mpl_toolkits.axes_grid1 import make_axes_locatable
import mngs

def conf_mat(
    ax,
    data,
    x_labels=None,
    y_labels=None,
    title="Confusion Matrix",
    cmap="Blues",
    cbar=True,
    cbar_kw={},
    label_rotation_xy=(15, 15),
    x_extend_ratio=1.0,
    y_extend_ratio=1.0,
    bacc=False,
    **kwargs
):
    """
    Create a confusion matrix heatmap from a numpy array or pandas DataFrame.

    Parameters
    ----------
    ax : matplotlib.axes.Axes
        The axes object to plot on
    data : array-like or pandas.DataFrame
        The confusion matrix data
    x_labels : list, optional
        Labels for x-axis
    y_labels : list, optional
        Labels for y-axis
    title : str, optional
        Title of the plot
    cmap : str, optional
        Colormap for the heatmap
    cbar : bool, optional
        Whether to display a colorbar
    cbar_kw : dict, optional
        Additional arguments for colorbar
    label_rotation_xy : tuple, optional
        Rotation angles for x and y labels
    x_extend_ratio : float, optional
        Ratio to extend x-axis
    y_extend_ratio : float, optional
        Ratio to extend y-axis
    bacc : bool, optional
        Whether to return the balanced accuracy
    **kwargs : dict
        Additional arguments for sns.heatmap

    Returns
    -------
    ax : matplotlib.axes.Axes
        The axes object with the confusion matrix plotted
    bacc : float, optional
        The balanced accuracy (if return_bacc is True)

    Example
    -------
    data = np.array([[10, 2, 0], [1, 15, 3], [0, 2, 20]])
    fig, ax = plt.subplots()
    ax = conf_mat(ax, data, x_labels=['A', 'B', 'C'], y_labels=['X', 'Y', 'Z'])
    plt.show()
    """

    fig = ax.fig

    if not isinstance(data, pd.DataFrame):
        data = pd.DataFrame(data)

    bacc_val = calc_bACC_from_cm(data.values)
    title = f"{title} (bACC = {bacc_val:.3f})"

    res = sns.heatmap(
        data,
        ax=ax,
        cmap=cmap,
        annot=True,
        fmt=",d",
        cbar=False,
        vmin=0,
        **kwargs
    )

    res.invert_yaxis()

    for _, spine in res.spines.items():
        spine.set_visible(False)

    ax.set_xlabel("Predicted label")
    ax.set_ylabel("True label")
    ax.set_title(title)

    if x_labels is not None:
        ax.set_xticklabels(x_labels)
    if y_labels is not None:
        ax.set_yticklabels(y_labels)

    ax = mngs.plt.ax.extend(ax, x_extend_ratio, y_extend_ratio)
    if data.shape[0] == data.shape[1]:
        ax.set_box_aspect(1)
        ax.set_xticklabels(
            ax.get_xticklabels(),
            rotation=label_rotation_xy[0],
            fontdict={"verticalalignment": "top"},
        )
        ax.set_yticklabels(
            ax.get_yticklabels(),
            rotation=label_rotation_xy[1],
            fontdict={"horizontalalignment": "right"},
        )

    # if cbar:
    #     divider = make_axes_locatable(ax)
    #     cax = divider.append_axes("right", size="5%", pad=0.1)
    #     bbox = ax.get_position()
    #     width_orig = bbox.x1 - bbox.x0
    #     width_tgt = width_orig * x_extend_ratio
    #     dx = width_orig - width_tgt
    #     cax = mngs.plt.ax.shift(cax, dx=-dx * 2.54, dy=0)
    #     fig.add_axes(cax)

    #     vmax = data.values.max().astype(int)
    #     norm = matplotlib.colors.Normalize(vmin=0, vmax=vmax)
    #     cbar = fig.colorbar(
    #         plt.cm.ScalarMappable(norm=norm, cmap=cmap),
    #         cax=cax,
    #         **cbar_kw
    #     )
    #     cbar.locator = matplotlib.ticker.MaxNLocator(nbins=4)
    #     cbar.update_ticks()
    #     cbar.outline.set_edgecolor("white")

    if bacc:
        return ax, bacc_val
    return ax

def calc_bACC_from_cm(confusion_matrix):
    """
    Calculate balanced accuracy from a confusion matrix.

    Parameters
    ----------
    confusion_matrix : array-like
        The confusion matrix

    Returns
    -------
    float
        The balanced accuracy

    Example
    -------
    cm = np.array([[10, 2, 0], [1, 15, 3], [0, 2, 20]])
    bacc = calc_bACC_from_cm(cm)
    print(f"Balanced Accuracy: {bacc:.3f}")
    """
    with mngs.gen.suppress_output():
        try:
            per_class = np.diag(confusion_matrix) / np.nansum(confusion_matrix, axis=1)
            bacc = np.nanmean(per_class)
        except:
            bacc = np.nan
        return round(bacc, 3)

def main():
    data = np.array([[10, 2, 0], [1, 15, 3], [0, 2, 20]])
    fig, ax = mngs.plt.subplots()
    ax, bacc = conf_mat(ax, data, x_labels=['A', 'B', 'C'], y_labels=['X', 'Y', 'Z'], bacc=True)
    print(bacc)
    mngs.io.save(fig, "fig.jpg")

if __name__ == '__main__':
    CONFIG, sys.stdout, sys.stderr, plt, CC = mngs.gen.start(sys, plt, verbose=False, agg=True)
    main()
    mngs.gen.close(CONFIG, verbose=False, notify=False)

# EOF



#!./env/bin/python3
# -*- coding: utf-8 -*-
# Time-stamp: "2024-07-05 20:06:17 (ywatanabe)"
# /home/ywatanabe/proj/mngs_repo/src/mngs/plt/ax/_ecdf.py

import mngs
import numpy as np
import pandas as pd

# def ecdf(ax, data):
#     data = np.hstack(data)
#     data = data[~np.isnan(data)]
#     nn = len(data)

#     data_sorted = np.sort(data)
#     ecdf = np.arange(1, len(data_sorted) + 1) / len(data_sorted)

#     ax.step(data_sorted, ecdf, where="post")
#     ax.plot(data_sorted, ecdf, marker=".", linestyle="none")

#     df = pd.DataFrame(
#         {
#             "x": data_sorted,
#             "y": ecdf,
#             "n": nn,
#         }
#     )
#     return ax, df


def ecdf(ax, data, **kwargs):
    # Flatten and remove NaN values
    data = np.hstack(data)
    data = data[~np.isnan(data)]
    nn = len(data)

    # Sort the data and compute the ECDF values
    data_sorted = np.sort(data)
    ecdf_perc = 100 * np.arange(1, len(data_sorted) + 1) / len(data_sorted)

    # Create the pseudo x-axis for step plotting
    x_step = np.repeat(data_sorted, 2)[1:]
    y_step = np.repeat(ecdf_perc, 2)[:-1]

    # Plot the ECDF_PERC using steps
    ax.plot(x_step, y_step, drawstyle="steps-post", **kwargs)

    # Scatter the original data points
    ax.plot(data_sorted, ecdf_perc, marker=".", linestyle="none")

    # Set ylim, xlim, and aspect ratio
    ax.set_ylim(0, 100)  # Set y-axis limits
    ax.set_xlim(0, 1.0)  # Set x-axis limits
    # ax.set_aspect(1.0)  # Set aspect ratio

    # Create a DataFrame to hold the ECDF_PERC data
    df = mngs.pd.force_df(
        {
            "x": data_sorted,
            "y": ecdf_perc,
            "n": nn,
            "x_step": x_step,
            "y_step": y_step,
        }
    )

    # Return the matplotlib axis and DataFrame
    return ax, df


# EOF



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Time-stamp: "2024-10-07 22:59:12 (ywatanabe)"

from bisect import bisect_right
import numpy as np
import mngs
from scipy import stats
from typing import Any, Literal, Dict, Callable
from mngs.gen import numpy_fn
import multiprocessing as mp
from functools import partial
import pandas as pd
# def _compute_surrogate(args):
#     data1, data2, corr_func, seed = args
#     np.random.seed(seed)

#     data1 = data1
#     data2 = np.random.permutation(data2)

#     non_nan_indices = ~(np.isnan(data1) | np.isnan(data2))
#     data1, data2 = data1[non_nan_indices], data2[non_nan_indices]

#     return corr_func(data1, data2)[0]
def _compute_surrogate(args):
    data1, data2, corr_func, seed = args
    np.random.seed(seed)

    data2 = np.random.permutation(data2)

    if corr_func == stats.spearmanr:
        return corr_func(data1, data2)[0]
    else:
        non_nan_indices = ~(np.isnan(data1) | np.isnan(data2))
        return corr_func(data1[non_nan_indices], data2[non_nan_indices])[0]

def _corr_test_base(
    data1: np.ndarray,
    data2: np.ndarray,
    only_significant: bool,
    n_perm: int,
    seed: int,
    corr_func: Callable,
    test_name: str,
    n_jobs: int = -1
) -> Dict[str, Any]:
    np.random.seed(seed)
    seeds = np.random.randint(0, n_perm*100, size=n_perm)

    if corr_func != stats.spearmanr:
        # Convert to numeric, replacing non-numeric values with NaN
        data1 = pd.to_numeric(data1, errors='coerce')
        data2 = pd.to_numeric(data2, errors='coerce')

        # Remove NaN values
        mask = ~(np.isnan(data1) | np.isnan(data2))
        data1 = data1[mask]
        data2 = data2[mask]

    if len(data1) < 2 or len(data2) < 2:
        raise ValueError("Not enough valid numeric data points for correlation.")

    corr_obs, _ = corr_func(data1, data2)

    # def _corr_test_base(
    #     data1: np.ndarray,
    #     data2: np.ndarray,
    #     only_significant: bool,
    #     n_perm: int,
    #     seed: int,
    #     corr_func: Callable,
    #     test_name: str,
    #     n_jobs: int = -1
    # ) -> Dict[str, Any]:
    #     # Random seed
    #     np.random.seed(seed)
    #     seeds = np.random.randint(0, n_perm*100, size=n_perm)

    #     # Convert data to numeric type and handle non-numeric values
    #     data1 = np.asarray(data1, dtype=float)
    #     data2 = np.asarray(data2, dtype=float)

    #     # Remove NaN values
    #     mask = ~(np.isnan(data1) | np.isnan(data2))
    #     data1 = data1[mask]
    #     data2 = data2[mask]

    #     corr_obs, _ = corr_func(data1, data2)

    if n_jobs != 1:
        n_jobs = mp.cpu_count() if n_jobs == -1 else n_jobs
        with mp.Pool(n_jobs) as pool:
            surrogate = np.array(pool.map(_compute_surrogate,
                                          [(data1, data2, corr_func, s) for s in seeds]))
    else:
        surrogate = np.array([
            _compute_surrogate((data1, data2, corr_func, s))
            for s in seeds
        ])

    # Add normality test for surrogate distribution
    _, p_normal = stats.normaltest(surrogate)
    if not p_normal > 0.05:
        print(f"Warning: Surrogate distribution may not be normal (p={p_normal:.3f})")


    rank = bisect_right(sorted(surrogate), corr_obs)
    pvalue = min(rank, n_perm - rank) / n_perm * 2

    stars = mngs.stats.p2stars(pvalue)
    sample_size = len(data1)
    effect_size = np.abs(corr_obs)

    result_string = (
        f"{test_name} Corr. = {corr_obs:.3f}; p-value = {pvalue:.3f} "
        f"(n={sample_size:,}, eff={effect_size:.3f}) {stars}"
    )

    if not only_significant or (only_significant and pvalue < 0.05):
        print(result_string)

    return {
        "p_value": round(pvalue, 3),
        "stars": stars,
        "effsize": round(effect_size, 3),
        "corr": round(corr_obs, 3),
        "surrogate": surrogate,
        "n": sample_size,
        "test_name": f"Permutation-based {test_name} correlation test",
        "statistic": round(corr_obs, 3),
        "H0": f"There is no {test_name.lower()} correlation between the two variables",
    }

@numpy_fn
def corr_test_spearman(
    data1: np.ndarray,
    data2: np.ndarray,
    only_significant: bool = False,
    n_perm: int = 1_000,
    seed: int = 42,
) -> Dict[str, Any]:
    return _corr_test_base(data1, data2, only_significant, n_perm, seed, stats.spearmanr, "Spearman")

@numpy_fn
def corr_test_pearson(
    data1: np.ndarray,
    data2: np.ndarray,
    only_significant: bool = False,
    n_perm: int = 1_000,
    seed: int = 42,
) -> Dict[str, Any]:
    return _corr_test_base(data1, data2, only_significant, n_perm, seed, stats.pearsonr, "Pearson")

@numpy_fn
def corr_test(
    data1: np.ndarray,
    data2: np.ndarray,
    test: Literal["pearson", "spearman"] = "pearson",
    only_significant: bool = False,
    n_perm: int = 1_000,
    seed: int = 42,
) -> Dict[str, Any]:
    """
    Performs a correlation test between two datasets using permutation.

    Parameters
    ----------
    data1 : np.ndarray
        First dataset for correlation.
    data2 : np.ndarray
        Second dataset for correlation.
    test : {"pearson", "spearman"}, optional
        Type of correlation test to perform. Default is "pearson".
    only_significant : bool, optional
        If True, only prints significant results. Default is False.
    n_perm : int, optional
        Number of permutations for the test. Default is 1,000.
    seed : int, optional
        Random seed for reproducibility. Default is 42.

    Returns
    -------
    Dict[str, Any]
        Contains 'p_value', 'stars', 'effsize', 'corr', 'surrogate', 'n', 'test_name', 'statistic', and 'H0'.

    Example
    -------
    >>> xx = np.array([3, 4, 4, 5, 7, 8, 10, 12, 13, 15])
    >>> yy = np.array([2, 4, 4, 5, 4, 7, 8, 19, 14, 10])
    >>> results = corr_test(xx, yy, test="pearson")
    """
    if test == "spearman":
        return corr_test_spearman(data1, data2, only_significant, n_perm, seed)
    elif test == "pearson":
        return corr_test_pearson(data1, data2, only_significant, n_perm, seed)
    else:
        raise ValueError("Invalid test type. Choose 'pearson' or 'spearman'.")

if __name__ == "__main__":
    xx = np.array([3, 4, 4, 5, 7, 8, 10, 12, 13, 15])
    yy = np.array([2, 4, 4, 5, 4, 7, 8, 19, 14, 10])
    results_pearson = corr_test(xx, yy)
    results_spearman = corr_test(xx, yy, test="spearman")
    print("Pearson results:", results_pearson)
    print("Spearman results:", results_spearman)
# #!/usr/bin/env python3
# # -*- coding: utf-8 -*-
# # Time-stamp: "2024-10-06 12:02:18 (ywatanabe)"

# from bisect import bisect_right
# import numpy as np
# import mngs
# from scipy import stats
# from typing import Any, Literal, Dict, List

# def _corr_test_base(
#     data1: np.ndarray,
#     data2: np.ndarray,
#     only_significant: bool,
#     n_perm: int,
#     seed: int,
#     corr_func: callable,
#     test_name: str
# ) -> Dict[str, Any]:
#     np.random.seed(seed)

#     non_nan_indices = ~(np.isnan(data1) | np.isnan(data2))
#     data1, data2 = data1[non_nan_indices], data2[non_nan_indices]

#     corr_obs, _ = corr_func(data1, data2)
#     surrogate = [
#         corr_func(data1, np.random.permutation(data2))[0]
#         for _ in range(n_perm)
#     ]

#     rank = bisect_right(sorted(surrogate), corr_obs)
#     pvalue = min(rank, n_perm - rank) / n_perm * 2

#     stars = mngs.stats.p2stars(pvalue)
#     sample_size = len(data1)
#     effect_size = np.abs(corr_obs)

#     # result_string = (
#     #     f"{test_name} Corr. = {corr_obs:.3f}; p-value = {pvalue:.3f} "
#     #     f"(n={sample_size:,}, eff={effect_size:.3f}) {stars}"
#     # )

#     if not only_significant or (only_significant and pvalue < 0.05):
#         print(result_string)

#     return {
#         "p_value": round(pvalue, 3),
#         "stars": stars,
#         "effsize": round(effect_size, 3),
#         "corr": round(corr_obs, 3),
#         "surrogate": np.array(surrogate),
#         "n": sample_size,
#         "test_name": f"Permutation-based {test_name} correlation",
#         "statistic": round(corr_obs, 3),
#         "H0": f"There is no {test_name.lower()} correlation between the two variables",
#     }

# def corr_test_spearman(
#     data1: np.ndarray,
#     data2: np.ndarray,
#     only_significant: bool = False,
#     n_perm: int = 1_000,
#     seed: int = 42,
# ) -> Dict[str, Any]:
#     return _corr_test_base(data1, data2, only_significant, n_perm, seed, stats.spearmanr, "Spearman")

# def corr_test_pearson(
#     data1: np.ndarray,
#     data2: np.ndarray,
#     only_significant: bool = False,
#     n_perm: int = 1_000,
#     seed: int = 42,
# ) -> Dict[str, Any]:
#     return _corr_test_base(data1, data2, only_significant, n_perm, seed, stats.pearsonr, "Pearson")

# def corr_test(
#     data1: np.ndarray,
#     data2: np.ndarray,
#     test: Literal["pearson", "spearman"] = "pearson",
#     only_significant: bool = False,
#     n_perm: int = 1_000,
#     seed: int = 42,
# ) -> Dict[str, Any]:
#     """
#     Performs a correlation test between two datasets using permutation.

#     Parameters
#     ----------
#     data1 : np.ndarray
#         First dataset for correlation.
#     data2 : np.ndarray
#         Second dataset for correlation.
#     test : {"pearson", "spearman"}, optional
#         Type of correlation test to perform. Default is "pearson".
#     only_significant : bool, optional
#         If True, only prints significant results. Default is False.
#     n_perm : int, optional
#         Number of permutations for the test. Default is 1,000.
#     seed : int, optional
#         Random seed for reproducibility. Default is 42.

#     Returns
#     -------
#     Dict[str, Any]
#         Contains 'p_value', 'stars', 'effsize', 'corr', 'surrogate', 'n', 'test_name', 'statistic', and 'H0'.

#     Example
#     -------
#     >>> xx = np.array([3, 4, 4, 5, 7, 8, 10, 12, 13, 15])
#     >>> yy = np.array([2, 4, 4, 5, 4, 7, 8, 19, 14, 10])
#     >>> results = corr_test(xx, yy, test="pearson")
#     """
#     if test == "spearman":
#         return corr_test_spearman(data1, data2, only_significant, n_perm, seed)
#     elif test == "pearson":
#         return corr_test_pearson(data1, data2, only_significant, n_perm, seed)
#     else:
#         raise ValueError("Invalid test type. Choose 'spearman' or 'pearson'.")

# if __name__ == "__main__":
#     xx = np.array([3, 4, 4, 5, 7, 8, 10, 12, 13, 15])
#     yy = np.array([2, 4, 4, 5, 4, 7, 8, 19, 14, 10])
#     results_spearman = corr_test(xx, yy)
#     results_pearson = corr_test(xx, yy, test="pearson")
#     print("Spearman results:", results_spearman)
#     print("Pearson results:", results_pearson)



#!./env/bin/python3
# -*- coding: utf-8 -*-
# Time-stamp: "2024-10-07 12:03:29 (ywatanabe)"
# /home/ywatanabe/proj/mngs_repo/src/mngs/pd/_merge_cols.py


"""
This script does XYZ.
"""

#!/usr/bin/env python3

import mngs
import numpy as np
import pandas as pd


################################################################################
## Pandas
################################################################################
def merge_columns(df, *args, sep1="_", sep2="-", name="merged"):
    """
    Join specified columns with their labels.

    Example:
        import pandas as pd
        import numpy as np

        df = pd.DataFrame(
            data=np.arange(25).reshape(5, 5),
            columns=["A", "B", "C", "D", "E"],
        )

        df1 = merge_columns(df, "A", "B", sep1="_", sep2="-")
        df2 = merge_columns(df, ["A", "B"], sep1="_", sep2="-")
        assert (df1 == df2).all().all() # True

        #     A   B   C   D   E        A_B
        # 0   0   1   2   3   4    A-0_B-1
        # 1   5   6   7   8   9    A-5_B-6
        # 2  10  11  12  13  14  A-10_B-11
        # 3  15  16  17  18  19  A-15_B-16
        # 4  20  21  22  23  24  A-20_B-21


    Parameters
    ----------
    df : pandas.DataFrame
        Input DataFrame
    *args : str or list
        Column names to join, either as separate arguments or a single list
    sep1 : str, optional
        Separator for joining column names, default "_"
    sep2 : str, optional
        Separator between column name and value, default "-"

    Returns
    -------
    pandas.DataFrame
        DataFrame with added merged column
    """
    _df = df.copy()
    columns = (
        args[0]
        if len(args) == 1 and isinstance(args[0], (list, tuple))
        else args
    )
    merged_col = _df[list(columns)].apply(
        lambda row: sep1.join(f"{col}{sep2}{val}" for col, val in row.items()),
        axis=1,
    )

    new_col_name = sep1.join(columns) if not name else str(name)
    _df[new_col_name] = merged_col
    return _df


merge_cols = merge_columns

# def merge_columns(_df, *columns):
#     import numpy as np
#     import pandas as pd

#     df = pd.DataFrame(
#         data=np.arange(25).reshape(5, 5),
#         columns=["A", "B", "C", "D", "E"],
#     )

#     columns = ["A", "B"]

#     df[columns].astype(str).apply("_".join, axis=1)
#     # 0      0_1
#     # 1      5_6
#     # 2    10_11
#     # 3    15_16
#     # 4    20_21

#     # How can I join like this?
#     # # 0      A_0-B_1
#     # # 1      A_5-B_6
#     # ...


# def merge_columns(_df, *columns):
#     """
#     Add merged columns in string.

#     Example:
#         import pandas as pd
#         import numpy as np

#         df = pd.DataFrame(data=np.arange(25).reshape(5,5),
#                           columns=["A", "B", "C", "D", "E"],
#         )

#         print(df)

#         # A   B   C   D   E
#         # 0   0   1   2   3   4
#         # 1   5   6   7   8   9
#         # 2  10  11  12  13  14
#         # 3  15  16  17  18  19
#         # 4  20  21  22  23  24

#         print(merge_columns(df, "A", "B", "C"))

#         #     A   B   C   D   E     A_B_C
#         # 0   0   1   2   3   4     0_1_2
#         # 1   5   6   7   8   9     5_6_7
#         # 2  10  11  12  13  14  10_11_12
#         # 3  15  16  17  18  19  15_16_17
#         # 4  20  21  22  23  24  20_21_22
#     """
#     from copy import deepcopy

#     df = deepcopy(_df)
#     merged = deepcopy(df[columns[0]])  # initialization
#     for c in columns[1:]:
#         merged = mngs.ml.utils.merge_labels(list(merged), deepcopy(df[c]))
#     df.loc[:, mngs.general.connect_strs(columns)] = merged
#     return df


# """
# Imports
# """
# import os
# import re
# import sys

# import matplotlib
# import matplotlib.pyplot as plt
# import seaborn as sns
# import importlib

# import mngs

# importlib.reload(mngs)

# import numpy as np
# import pandas as pd
# import torch
# import torch.nn as nn
# import torch.nn.functional as F
# from icecream import ic
# from natsort import natsorted
# from glob import glob
# from pprint import pprint
# import warnings
# import logging
# from tqdm import tqdm
# import xarray as xr

# # sys.path = ["."] + sys.path
# # from scripts import utils, load

# """
# Warnings
# """
# # warnings.simplefilter("ignore", UserWarning)


# """
# Config
# """
# # CONFIG = mngs.gen.load_configs()


# """
# Functions & Classes
# """


# if __name__ == "__main__":
#     # # Argument Parser
#     # import argparse
#     # parser = argparse.ArgumentParser(description='')
#     # parser.add_argument('--var', '-v', type=int, default=1, help='')
#     # parser.add_argument('--flag', '-f', action='store_true', default=False, help='')
#     # args = parser.parse_args()

#     # Main
#     CONFIG, sys.stdout, sys.stderr, plt, CC = mngs.gen.start(
#         sys, plt, verbose=False
#     )
#     main()
#     mngs.gen.close(CONFIG, verbose=False, notify=False)

# # EOF



#!./env/bin/python3
# -*- coding: utf-8 -*-
# Time-stamp: "2024-08-29 23:08:35 (ywatanabe)"
# /home/ywatanabe/proj/mngs_repo/src/mngs/pd/_replace.py


def replace(dataframe, replace_dict, regex=False, cols=None):
    """
    Replace values in specified columns of a DataFrame using a dictionary.

    Example
    -------
    import pandas as pd
    df = pd.DataFrame({'A': ['abc-123', 'def-456'], 'B': ['ghi-789', 'jkl-012']})
    replace_dict = {'-': '_', '1': 'one'}
    df_replaced = replace(df, replace_dict, cols=['A'])
    print(df_replaced)

    Parameters
    ----------
    dataframe : pandas.DataFrame
        Input DataFrame to modify.
    replace_dict : dict
        Dictionary of old values (keys) and new values (values) to replace.
    regex : bool, optional
        If True, treat `replace_dict` keys as regular expressions. Default is False.
    cols : list of str, optional
        List of column names to apply replacements. If None, apply to all string columns.

    Returns
    -------
    pandas.DataFrame
        DataFrame with specified replacements applied.
    """

    dataframe = dataframe.copy()

    if cols is None:
        cols = dataframe.select_dtypes(include=["object"]).columns

    for column in cols:
        if dataframe[column].dtype == "object":
            for old_value, new_value in replace_dict.items():
                dataframe[column] = dataframe[column].str.replace(
                    old_value, new_value, regex=regex
                )

    return dataframe



#!./env/bin/python3
# -*- coding: utf-8 -*-
# Time-stamp: "2024-04-22 23:26:53"
# Author: Yusuke Watanabe (ywata1989@gmail.com)

"""
This script does XYZ.
"""

# Functions
import os
import subprocess
import sys
from datetime import datetime

import matplotlib.pyplot as plt
import mngs
import pandas as pd
import psutil


# Functions
def get_proc_usages():
    """
    Retrieves the current usage statistics for the CPU, RAM, GPU, and VRAM.

    This function fetches the current usage percentages for the CPU and GPU, as well as the current usage in GiB for RAM and VRAM.
    The data is then compiled into a pandas DataFrame with the current timestamp.

    Returns:
        pd.DataFrame: A pandas DataFrame containing the current usage statistics with the following columns:
                      - Time: The timestamp when the data was retrieved.
                      - CPU [%]: The CPU usage percentage.
                      - RAM [GiB]: The RAM usage in GiB.
                      - GPU [%]: The GPU usage percentage.
                      - VRAM [GiB]: The VRAM usage in GiB.
                      Each row in the DataFrame represents a single instance of data retrieval, rounded to 1 decimal place.

    Example:
        >>> usage_df = get_proc_usages()
        >>> print(usage_df)
    """
    cpu_perc, ram_gb = _get_cpu_usage()
    gpu_perc, vram_gb = _get_gpu_usage()

    sr = pd.Series(
        {
            "Time": datetime.now(),
            "CPU [%]": cpu_perc,
            "RAM [GiB]": ram_gb,
            "GPU [%]": gpu_perc,
            "VRAM [GiB]": vram_gb,
        }
    )

    df = pd.DataFrame(sr).round(1).T

    return df


def _get_cpu_usage(process=os.getpid(), n_round=1):
    cpu_usage_perc = psutil.cpu_percent()
    ram_usage_gb = (
        psutil.virtual_memory().percent
        / 100
        * psutil.virtual_memory().total
        / (1024**3)
    )
    return round(cpu_usage_perc, n_round), round(ram_usage_gb, n_round)


def _get_gpu_usage(n_round=1):
    result = subprocess.run(
        [
            "nvidia-smi",
            "--query-gpu=utilization.gpu,memory.used",
            "--format=csv,nounits,noheader",
        ],
        capture_output=True,
        text=True,
    )
    gpu_usage_perc, _vram_usage_mib = result.stdout.strip().split(",")
    vram_usage_gb = float(_vram_usage_mib) / 1024
    return round(float(gpu_usage_perc), n_round), round(
        float(vram_usage_gb), n_round
    )


# (YOUR AWESOME CODE)

if __name__ == "__main__":
    # Start
    CONFIG, sys.stdout, sys.stderr, plt, CC = mngs.gen.start(
        sys, plt, verbose=False
    )

    usage = mngs.res.get_proc_usages()
    mngs.io.save(usage, "usage.csv")

    # Close
    mngs.gen.close(CONFIG, verbose=False, notify=False)

# EOF

"""
/home/ywatanabe/proj/entrance/mngs/res/_get_procs_usage.py
"""



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Time-stamp: "2024-10-13 01:05:27 (ywatanabe)"
# File: /home/ywatanabe/proj/_mngs_repo_openhands/src/mngs/io/_save.py


"""
1. Functionality:
   - Provides utilities for saving various data types to different file formats.
2. Input:
   - Objects to be saved (e.g., NumPy arrays, PyTorch tensors, Pandas DataFrames, etc.)
   - File path or name where the object should be saved
3. Output:
   - Saved files in various formats (e.g., CSV, NPY, PKL, JOBLIB, PNG, HTML, TIFF, MP4, YAML, JSON, HDF5, PTH, MAT, CBM)
4. Prerequisites:
   - Python 3.x
   - Required libraries: numpy, pandas, torch, matplotlib, plotly, h5py, joblib, PIL, ruamel.yaml
"""

"""Imports"""
from typing import List, Tuple, Dict, Any, Union, Sequence, Literal
from collections.abc import Iterable
import numpy as np
import csv
import inspect
import io as _io
import json
import os
import pickle
import shutil
from shutil import move

import h5py
import joblib
import mngs
import numpy as np
import pandas as pd
import plotly
import scipy
import torch
from matplotlib import animation
from PIL import Image
from ruamel.yaml import YAML
import xarray as xr
import logging

ArrayLike = Union[
    List,
    Tuple,
    np.ndarray,
    pd.Series,
    pd.DataFrame,
    xr.DataArray,
    torch.Tensor,
]

# def save(
#     obj,
#     sfname_or_spath,
#     makedirs=True,
#     verbose=True,
#     from_cwd=False,
#     dry_run=False,
#     **kwargs,
# ):
#     """
#     Saves an object to a file with the specified format, determined by the file extension.
#     The function supports saving data in various formats including CSV, NPY, PKL, JOBLIB, PNG, HTML, TIFF, MP4, YAML, JSON, HDF5, PTH, MAT, and CBM.

#     Arguments:
#         obj (object): The object to be saved. Can be a NumPy array, PyTorch tensor, Pandas DataFrame, or any serializable object depending on the file format.
#         sfname_or_spath (str): The file name or path where the object should be saved. The file extension determines the format.
#         makedirs (bool, optional): If True, the function will create the directory path if it does not exist. Defaults to True.
#         verbose (bool, optional): If True, prints a message upon successful saving. Defaults to True.
#         **kwargs: Additional keyword arguments to pass to the underlying save function of the specific format.

#     Returns:
#         None. The function saves the object to a file and does not return any value.

#     Supported Formats:
#         - .csv: Pandas DataFrames or listed scalars.
#         - .npy: NumPy arrays.
#         - .npz: NumPy arrays.
#         - .pkl: Serializable Python objects using pickle.
#         - .joblib: Objects using joblib with compression.
#         - .png: Images, including Plotly figures and Matplotlib plots.
#         - .tiff / .tif: Images in TIFF format, typically from Matplotlib plots.
#         - .jpeg / .jpg: Images in JPEG format, typically from Matplotlib plots.
#         - .svg : Images in the SVG format, typically from Matplotlib plots.
#         - .html: Plotly figures as HTML files.
#         - .mp4: Videos, likely from animations created with Matplotlib.
#         - .yaml: Data in YAML format.
#         - .json: Data in JSON format.
#         - .hdf5: Data in HDF5 format, useful for large datasets.
#         - .pth: PyTorch model states.
#         - .mat: Data in MATLAB format using `scipy.io.savemat`.
#         - .cbm: CatBoost models.

#     Note:
#         - The function dynamically selects the appropriate saving mechanism based on the file extension.
#         - Ensure that the object type is compatible with the chosen file format.
#         - For custom saving mechanisms or unsupported formats, consider extending the function or using the specific library's save function directly.

#     Example:
#         import mngs

#         import numpy as np
#         import pandas as pd
#         import torch
#         import matplotlib.pyplot as plt

#         # .npy
#         arr = np.array([1, 2, 3])
#         mngs.io.save(arr, "xxx.npy")
#         # arr = mngs.io.load("xxx.npy")

#         # .csv
#         df = pd.DataFrame({"col1": [1, 2], "col2": [3, 4]})
#         mngs.io.save(df, "xxx.csv")
#         # df = mngs.io.load("xxx.csv")

#         # .pth
#         tensor = torch.tensor([1, 2, 3])
#         mngs.io.save(obj, "xxx.pth")
#         # tensor = mngs.io.load("xxx.pth")

#         # .pkl
#         _dict = {"a": 1, "b": 2, "c": [3, 4, 5]} # serializable object like dict, list, ...
#         mngs.io.save(_dict, "xxx.pkl")
#         # _dict = mngs.io.load("xxx.pkl")

#         # .png | .tiff | .jpg
#         plt.figure()
#         plt.plot(np.array([1, 2, 3]))
#         mngs.io.save(plt, "xxx.png") # "xxx.tiff", "xxx.tif", "xxx.jpeg, or "xxx.jpg"

#         # .yaml
#         mngs.io.save(_dict, "xxx.yaml")
#         # _dict = mngs.io.load("xxx.yaml")

#         # .json
#         mngs.io.save(_dict, "xxx.json")
#         # _dict = mngs.io.load("xxx.json")
#     """


def save(
    obj: Any,
    sfname_or_spath: str,
    makedirs: bool = True,
    verbose: bool = True,
    from_cwd: bool = False,
    dry_run: bool = False,
    **kwargs,
) -> None:
    """
    Save an object to a file with the specified format.

    Parameters
    ----------
    obj : Any
        The object to be saved. Can be a NumPy array, PyTorch tensor, Pandas DataFrame, or any serializable object.
    sfname_or_spath : str
        The file name or path where the object should be saved. The file extension determines the format.
    makedirs : bool, optional
        If True, create the directory path if it does not exist. Default is True.
    verbose : bool, optional
        If True, print a message upon successful saving. Default is True.
    from_cwd : bool, optional
        If True, create a symlink from the current working directory. Default is False.
    dry_run : bool, optional
        If True, simulate the saving process without actually writing files. Default is False.
    **kwargs
        Additional keyword arguments to pass to the underlying save function of the specific format.

    Returns
    -------
    None

    Notes
    -----
    Supported formats include CSV, NPY, PKL, JOBLIB, PNG, HTML, TIFF, MP4, YAML, JSON, HDF5, PTH, MAT, and CBM.
    The function dynamically selects the appropriate saving mechanism based on the file extension.

    Examples
    --------
    >>> import mngs
    >>> import numpy as np
    >>> import pandas as pd
    >>> import torch
    >>> import matplotlib.pyplot as plt

    >>> # Save NumPy array
    >>> arr = np.array([1, 2, 3])
    >>> mngs.io.save(arr, "data.npy")

    >>> # Save Pandas DataFrame
    >>> df = pd.DataFrame({"col1": [1, 2], "col2": [3, 4]})
    >>> mngs.io.save(df, "data.csv")

    >>> # Save PyTorch tensor
    >>> tensor = torch.tensor([1, 2, 3])
    >>> mngs.io.save(tensor, "model.pth")

    >>> # Save dictionary
    >>> data_dict = {"a": 1, "b": 2, "c": [3, 4, 5]}
    >>> mngs.io.save(data_dict, "data.pkl")

    >>> # Save matplotlib figure
    >>> plt.figure()
    >>> plt.plot(np.array([1, 2, 3]))
    >>> mngs.io.save(plt, "plot.png")

    >>> # Save as YAML
    >>> mngs.io.save(data_dict, "config.yaml")

    >>> # Save as JSON
    >>> mngs.io.save(data_dict, "data.json")
    """
    try:
        ########################################
        # Determines the save directory from the script.
        # When save() is called in a /path/to/script.py, data will be saved under /path/to/ directoy.
        # On the other hand, when save() is called in an ipython environment, data will be saved in /tmp/USERNAME/
        # This process should be in this function for the intended behavior of inspect.
        ########################################
        spath, sfname = None, None

        if sfname_or_spath.startswith('f"'):
            sfname_or_spath = eval(sfname_or_spath)

        if sfname_or_spath.startswith("/"):
            spath = sfname_or_spath

        else:
            fpath = inspect.stack()[1].filename

            if ("ipython" in fpath) or ("<stdin>" in fpath):
                fpath = f'/tmp/{os.getenv("USER")}.py'

            fdir, fname, _ = mngs.path.split(fpath)
            spath = fdir + fname + "/" + sfname_or_spath

        # Corrects the spath
        spath = format_spath(spath)
        ########################################

        # Potential path to symlink
        spath_cwd = os.getcwd() + "/" + sfname_or_spath
        spath_cwd = format_spath(spath_cwd)

        # Removes spath and spath_cwd to prevent potential circular links
        for path in [spath, spath_cwd]:
            mngs.sh(f"rm -f {path}", verbose=False)

        if dry_run:
            print(mngs.gen.ct(f"\n(dry run) Saved to: {spath}", c="yellow"))
            return

        # Makes directory
        if makedirs:
            os.makedirs(os.path.dirname(spath), exist_ok=True)

        _save(
            obj,
            spath,
            verbose=verbose,
            from_cwd=from_cwd,
            dry_run=dry_run,
            **kwargs,
        )
        symlink(spath, spath_cwd, from_cwd, verbose)

    except Exception as e:
        logging.error(
            f"Error occurred while saving: {str(e)}"
            f"Debug: Initial fpath = {inspect.stack()[1].filename}"
            f"Debug: Final fpath = {fpath}"
            f"Debug: fdir = {fdir}, fname = {fname}"
            f"Debug: Final spath = {spath}"
        )


def symlink(spath, spath_cwd, from_cwd, verbose):
    if from_cwd and (spath != spath_cwd):
        os.makedirs(os.path.dirname(spath_cwd), exist_ok=True)
        mngs.sh(f"rm -f {spath_cwd}", verbose=False)
        mngs.sh(f"ln -sfr {spath} {spath_cwd}", verbose=False)
        if verbose:
            print(
                mngs.gen.color_text(f"\n(Symlinked to: {spath_cwd})", "yellow")
            )


def _save(obj, spath, verbose=True, from_cwd=False, dry_run=False, **kwargs):
    # csv
    if spath.endswith(".csv"):
        if isinstance(obj, (pd.Series, pd.DataFrame)):  # Series or DataFrame
            obj.to_csv(spath, **kwargs)

        if mngs.gen.is_listed_X(obj, [int, float]):  # listed scalars
            _save_listed_scalars_as_csv(
                obj,
                spath,
                **kwargs,
            )
        if mngs.gen.is_listed_X(obj, pd.DataFrame):  # listed DataFrame
            _save_listed_dfs_as_csv(obj, spath, **kwargs)

    # numpy
    elif spath.endswith(".npy"):
        np.save(spath, obj)

    # numpy npz
    elif spath.endswith(".npz"):
        if isinstance(obj, dict):
            np.savez_compressed(spath, **obj)
        elif isinstance(obj, (list, tuple)) and all(
            isinstance(x, np.ndarray) for x in obj
        ):
            obj = {str(ii): obj[ii] for ii in range(len(obj))}
            np.savez_compressed(spath, **obj)
        else:
            raise ValueError(
                "For .npz files, obj must be a dict of arrays or a list/tuple of arrays."
            )
    # pkl
    elif spath.endswith(".pkl"):
        with open(spath, "wb") as s:  # 'w'
            pickle.dump(obj, s)

    # joblib
    elif spath.endswith(".joblib"):
        with open(spath, "wb") as s:  # 'w'
            joblib.dump(obj, s, compress=3)

    # html
    elif spath.endswith(".html"):
        # plotly
        if isinstance(obj, plotly.graph_objs.Figure):
            obj.write_html(file=spath)

    # image ----------------------------------------
    elif any(
        [
            spath.endswith(image_ext)
            for image_ext in [
                ".png",
                ".tiff",
                ".tif",
                ".jpeg",
                ".jpg",
                ".svc",
            ]
        ]
    ):
        _save_image(obj, spath, **kwargs)
        ext = os.path.splitext(spath)[1].lower()
        try:
            ext_wo_dot = ext.replace(".", "")
            save(
                obj.to_sigma(),
                spath.replace(ext_wo_dot, "csv"),
                from_cwd=from_cwd,
                dry_run=dry_run,
                **kwargs,
            )
        except Exception as e:
            print(e)

    # mp4
    elif spath.endswith(".mp4"):
        obj.save(
            spath, writer="ffmpeg", **kwargs
        )  # obj is matplotlib.animation.FuncAnimation
        del obj
        # _mk_mp4(obj, spath)  # obj is matplotlib.pyplot.figure object
        # del obj

    # yaml
    elif spath.endswith(".yaml"):
        yaml = YAML()
        yaml.preserve_quotes = True  # Optional: if you want to preserve quotes
        yaml.indent(
            mapping=4, sequence=4, offset=4
        )  # Optional: set indentation

        with open(spath, "w") as f:
            yaml.dump(obj, f)

    # json
    elif spath.endswith(".json"):
        with open(spath, "w") as f:
            json.dump(obj, f, indent=4)

    # hdf5
    elif spath.endswith(".hdf5"):
        name_list, obj_list = []
        for k, v in obj.items():
            name_list.append(k)
            obj_list.append(v)
        with h5py.File(spath, "w") as hf:
            for name, obj in zip(name_list, obj_list):
                hf.create_dataset(name, data=obj)
    # pth
    elif spath.endswith(".pth"):
        torch.save(obj, spath)

    # mat
    elif spath.endswith(".mat"):
        scipy.io.savemat(spath, obj)

    # catboost model
    elif spath.endswith(".cbm"):
        obj.save_model(spath)

    # Text
    elif any(
        spath.endswith(ext)
        for ext in [".txt", ".md", ".py", ".html", ".css", ".js"]
    ):
        _save_text(obj, spath)

    else:
        raise ValueError(f"Unsupported file format. {spath} was not saved.")

    if verbose:
        if os.path.exists(spath):
            file_size = mngs.path.file_size(spath)
            print(
                mngs.gen.ct(f"\nSaved to: {spath} ({file_size})", c="yellow")
            )


# def check_encoding(file_path):
#     from chardet.universaldetector import UniversalDetector

#     detector = UniversalDetector()
#     with open(file_path, mode="rb") as f:
#         for binary in f:
#             detector.feed(binary)
#             if detector.done:
#                 break
#     detector.close()
#     enc = detector.result["encoding"]
#     return enc


def format_spath(spath):
    return spath.replace("/./", "/").replace("//", "/").replace(" ", "_")


def _save_image(obj, spath, **kwargs):

    # png
    if spath.endswith(".png"):
        # plotly
        if isinstance(obj, plotly.graph_objs.Figure):
            obj.write_image(file=spath, format="png")
        # PIL image
        elif isinstance(obj, Image.Image):
            obj.save(spath)
        # matplotlib
        else:
            try:
                obj.savefig(spath)
            except:
                obj.figure.savefig(spath)
        del obj

    # tiff
    elif spath.endswith(".tiff") or spath.endswith(".tif"):
        # PIL image
        if isinstance(obj, Image.Image):
            obj.save(spath)
        # matplotlib
        else:
            try:
                obj.savefig(spath, dpi=300, format="tiff")
            except:
                obj.figure.savefig(spath, dpi=300, format="tiff")

        del obj

    # jpeg
    elif spath.endswith(".jpeg") or spath.endswith(".jpg"):
        buf = _io.BytesIO()

        # plotly
        if isinstance(obj, plotly.graph_objs.Figure):
            obj.write_image(
                buf, format="png"
            )  # Saving plotly figure to buffer as PNG
            buf.seek(0)
            im = Image.open(buf)
            img.convert("RGB").save(spath, "JPEG")
            buf.close()

        # PIL image
        elif isinstance(obj, Image.Image):
            obj.save(spath)

        # matplotlib
        else:
            try:
                obj.savefig(buf, format="png")
            except:
                obj.figure.savefig(buf, format="png")

            buf.seek(0)
            img = Image.open(buf)
            img.convert("RGB").save(spath, "JPEG")  # Convert to JPEG and save
            buf.close()
        del obj

    # SVG
    elif spath.endswith(".svg"):
        # Plotly
        if isinstance(obj, plotly.graph_objs.Figure):
            obj.write_image(file=spath, format="svg")
        # Matplotlib
        else:
            try:
                obj.savefig(spath, format="svg")
            except AttributeError:
                obj.figure.savefig(spath, format="svg")
        del obj


def _save_text(obj, spath):
    with open(spath, "w") as file:
        file.write(obj)


def _save_listed_scalars_as_csv(
    listed_scalars,
    spath_csv,
    column_name="_",
    indi_suffix=None,
    round=3,
    overwrite=False,
    verbose=False,
):
    """Puts to df and save it as csv"""

    if overwrite == True:
        _mv_to_tmp(spath_csv, L=2)
    indi_suffix = (
        np.arange(len(listed_scalars)) if indi_suffix is None else indi_suffix
    )
    df = pd.DataFrame(
        {"{}".format(column_name): listed_scalars}, index=indi_suffix
    ).round(round)
    df.to_csv(spath_csv)
    if verbose:
        print("\nSaved to: {}\n".format(spath_csv))


def _save_listed_dfs_as_csv(
    listed_dfs,
    spath_csv,
    indi_suffix=None,
    overwrite=False,
    verbose=False,
):
    """listed_dfs:
        [df1, df2, df3, ..., dfN]. They will be written vertically in the order.

    spath_csv:
        /hoge/fuga/foo.csv

    indi_suffix:
        At the left top cell on the output csv file, '{}'.format(indi_suffix[i])
        will be added, where i is the index of the df.On the other hand,
        when indi_suffix=None is passed, only '{}'.format(i) will be added.
    """
    import numpy as np

    if overwrite == True:
        _mv_to_tmp(spath_csv, L=2)

    indi_suffix = (
        np.arange(len(listed_dfs)) if indi_suffix is None else indi_suffix
    )
    for i, df in enumerate(listed_dfs):
        with open(spath_csv, mode="a") as f:
            f_writer = csv.writer(f)
            i_suffix = indi_suffix[i]
            f_writer.writerow(["{}".format(indi_suffix[i])])
        df.to_csv(spath_csv, mode="a", index=True, header=True)
        with open(spath_csv, mode="a") as f:
            f_writer = csv.writer(f)
            f_writer.writerow([""])
    if verbose:
        print("Saved to: {}".format(spath_csv))


def _mv_to_tmp(fpath, L=2):
    try:
        tgt_fname = mngs.gen.connect_strs(fpath.split("/")[-L:], filler="-")
        tgt_fpath = "/tmp/{}".format(tgt_fname)
        move(fpath, tgt_fpath)
        print("Moved to: {}".format(tgt_fpath))
    except:
        pass


def _mk_mp4(fig, spath_mp4):
    axes = fig.get_axes()

    def init():
        return (fig,)

    def animate(i):
        for ax in axes:
            ax.view_init(elev=10.0, azim=i)
        return (fig,)

    anim = animation.FuncAnimation(
        fig, animate, init_func=init, frames=360, interval=20, blit=True
    )

    writermp4 = animation.FFMpegWriter(
        fps=60, extra_args=["-vcodec", "libx264"]
    )
    anim.save(spath_mp4, writer=writermp4)
    print("\nSaving to: {}\n".format(spath_mp4))


def save_optuna_study_as_csv_and_pngs(study, sdir):
    import optuna

    ## Trials DataFrame
    trials_df = study.trials_dataframe()

    ## Figures
    hparams_keys = list(study.best_params.keys())
    slice_plot = optuna.visualization.plot_slice(study, params=hparams_keys)
    contour_plot = optuna.visualization.plot_contour(
        study, params=hparams_keys
    )
    optim_hist_plot = optuna.visualization.plot_optimization_history(study)
    parallel_coord_plot = optuna.visualization.plot_parallel_coordinate(
        study, params=hparams_keys
    )
    hparam_importances_plot = optuna.visualization.plot_param_importances(
        study
    )
    figs_dict = dict(
        slice_plot=slice_plot,
        contour_plot=contour_plot,
        optim_hist_plot=optim_hist_plot,
        parallel_coord_plot=parallel_coord_plot,
        hparam_importances_plot=hparam_importances_plot,
    )

    ## Saves
    save(trials_df, sdir + "trials_df.csv")

    for figname, fig in figs_dict.items():
        save(fig, sdir + f"{figname}.png")


def format_spath(spath):
    return spath.replace("/./", "/").replace("//", "/").replace(" ", "_")


if __name__ == "__main__":
    import mngs

    mngs.io.save("aaa", "text.txt")



#!/usr/bin/env python3

import mngs
import numpy as np

# y1, y2 = T_tra, M_tra
# def merge_labels(y1, y2):
#     y = [str(z1) + "-" + str(z2) for z1, z2 in zip(y1, y2)]
#     conv_d = {z: i for i, z in enumerate(np.unique(y))}
#     y = [conv_d[z] for z in y]
#     return y


def merge_labels(*ys, to_int=False):
    if not len(ys) > 1:  # Check if more than two arguments are passed
        return ys[0]
    else:
        y = [mngs.general.connect_nums(zs) for zs in zip(*ys)]
        if to_int:
            conv_d = {z: i for i, z in enumerate(np.unique(y))}
            y = [conv_d[z] for z in y]
        return np.array(y)



#!/usr/bin/env python


import matplotlib


def set_n_ticks(
    ax,
    n_xticks=4,
    n_yticks=4,
):
    """
    Example:
        ax = set_n_ticks(ax)
    """

    if n_xticks is not None:
        ax.xaxis.set_major_locator(matplotlib.ticker.MaxNLocator(n_xticks))

    if n_yticks is not None:
        ax.yaxis.set_major_locator(matplotlib.ticker.MaxNLocator(n_yticks))

    # Force the figure to redraw to reflect changes
    ax.figure.canvas.draw()

    return ax



#!./env/bin/python3
# -*- coding: utf-8 -*-
# Time-stamp: "2024-10-27 13:02:01 (ywatanabe)"
# /home/ywatanabe/proj/mngs/src/mngs/ml/chat.py


"""Imports"""
import os
import random

from ._Anthropic import Anthropic
from ._Google import Google
from ._Llama import Llama
from ._OpenAI import OpenAI
from ._Perplexity import Perplexity
from .PARAMS import MODELS
import pandas as pd

"""Functions & Classes"""

"""Parameters"""
def genai_factory(
    model="gpt-3.5-turbo",
    stream=False,
    api_key=None,
    seed=None,
    temperature=1.0,
    n_keep=1,
    chat_history=None,
    max_tokens=4096,
):
    """Factory function to create an instance of an AI model handler."""
    AVAILABLE_MODELS = MODELS.name.tolist()

    if model not in AVAILABLE_MODELS:
        raise ValueError(
            f'Model "{model}" is not available. Please choose from:{MODELS.name.tolist()}'
        )

    provider = MODELS[MODELS.name == model].provider.iloc[0]

    model_class = globals()[provider]

    # Select a random API key from the list
    if isinstance(api_key, (list, tuple)):
        api_key = random.choice(api_key)

    return model_class(
        model=model,
        stream=stream,
        api_key=api_key,
        seed=seed,
        temperature=temperature,
        n_keep=n_keep,
        chat_history=chat_history,
        max_tokens=max_tokens,
    )


def main(
    model="gpt-3.5-turbo",
    stream=False,
    prompt="Hi, please tell me about the hippocampus",
    seed=None,
    temperature=1.0,
):
    m = genai_factory(model, stream=stream, seed=seed, temperature=temperature)
    out = m(prompt)
    return out


################################################################################
# Helper functions
################################################################################
# def model2apikey(model):
#     """Retrieve the API key for a given model from environment variables."""
#     for config in MODEL_CONFIG.values():
#         if model in config["models"]:
#             api_key = os.getenv(config["api_key_env"])
#             if not api_key:
#                 raise EnvironmentError(
#                     f"API key for {model} not found in environment."
#                 )
#             return api_key
#     raise ValueError(f"Model {model} is not supported.")


def test_all(seed=None, temperature=1.0):
    model_names = [
        "claude-3-5-sonnet-20240620",
        # "gpt-4",
        # "claude-3-opus-20240229",
        # "gemini-pro",
        # "llama-3-sonar-large-32k-online",
    ]

    for model_name in model_names:
        for stream in [False, True]:
            model = GenAI(
                model_name, stream=stream, seed=seed, temperature=temperature
            )
            # prompt = "Hi. Tell me your name just within a line."
            prompt = "Hi. Tell me about the hippocampus."

            print(
                f"\n{'-'*40}\n{model.model}\nStream: {stream}\nSeed: {seed}\nTemperature: {temperature}\n{'-'*40}"
            )
            print(model(prompt))
            print(model.available_models)


def main(
    model="gemini-1.5-pro-latest",
    stream=False,
    prompt="Hi, please tell me about the hippocampus",
    seed=None,
    temperature=1.0,
):

    m = mngs.ai.GenAI(
        model=model,
        api_key=os.getenv("GOOGLE_API_KEY"),
        stream=stream,
        seed=seed,
        temperature=temperature,
    )
    out = m(prompt)

    return out


if __name__ == "__main__":
    import sys

    import matplotlib.pyplot as plt
    import mngs

    CONFIG, sys.stdout, sys.stderr, plt, CC = mngs.gen.start(
        sys, plt, verbose=False
    )
    test_all()

    mngs.gen.close(CONFIG, verbose=False, notify=False)

# EOF



#!/usr/bin/env python3


import inspect
import os

import mngs
import numpy as np

def this_path(when_ipython="/tmp/fake.py"):
    __file__ = inspect.stack()[1].filename
    if "ipython" in __file__:  # for ipython
        __file__ = when_ipython  # "/tmp/fake.py"
    return __file__


get_this_path = this_path


def spath(sfname=".", makedirs=False):

    __file__ = inspect.stack()[1].filename
    if "ipython" in __file__:  # for ipython
        __file__ = f'/tmp/fake-{os.getenv("USER")}.py'

    ## spath
    fpath = __file__
    fdir, fname, _ = split(fpath)
    sdir = fdir + fname + "/"
    spath = sdir + sfname

    if makedirs:
        os.makedirs(mngs.path.split(spath)[0], exist_ok=True)

    return spath


mk_spath = spath


def split(fpath):
    """Split a file path to (1) the directory path, (2) the file name, and (3) the file extention
    Example:
        dirname, fname, ext = split('../data/01/day1/split_octave/2kHz_mat/tt8-2.mat')
        print(dirname) # '../data/01/day1/split_octave/2kHz_mat/'
        print(fname) # 'tt8-2'
        print(ext) # '.mat'
    """
    dirname = os.path.dirname(fpath) + "/"
    base = os.path.basename(fpath)
    fname, ext = os.path.splitext(base)
    return dirname, fname, ext


def file_size(path):
    if os.path.exists(path):
        file_size_bytes = os.path.getsize(path)
        return mngs.gen.readable_bytes(file_size_bytes)
    else:
        return "(Not Found)"

def getsize(path):
    if os.path.exists(path):
        return os.path.getsize(path)
    else:
        return np.nan



#!./env/bin/python3
# -*- coding: utf-8 -*-
# Time-stamp: "2024-04-17 09:19:59"
# Author: Yusuke Watanabe (ywata1989@gmail.com)

"""
This script does XYZ.
"""

import os
import re
import sys
from glob import glob

import matplotlib.pyplot as plt

# Imports
import mngs


# Functions
def find_latest(dirname, fname, ext, version_prefix="_v"):
    version_pattern = re.compile(
        rf"({re.escape(fname)}{re.escape(version_prefix)})(\d+)({re.escape(ext)})$"
    )

    glob_pattern = os.path.join(dirname, f"{fname}{version_prefix}*{ext}")
    files = glob(glob_pattern)

    highest_version = 0
    latest_file = None

    for file in files:
        filename = os.path.basename(file)
        match = version_pattern.search(filename)
        if match:
            version_num = int(match.group(2))
            if version_num > highest_version:
                highest_version = version_num
                latest_file = file

    return latest_file


## Version
def increment_version(dirname, fname, ext, version_prefix="_v"):
    # Create a regex pattern to match the version number in the filename
    version_pattern = re.compile(
        rf"({re.escape(fname)}{re.escape(version_prefix)})(\d+)({re.escape(ext)})$"
    )

    # Construct the glob pattern to find all files that match the pattern
    glob_pattern = os.path.join(dirname, f"{fname}{version_prefix}*{ext}")

    # Use glob to find all files that match the pattern
    files = glob(glob_pattern)

    # Initialize the highest version number
    highest_version = 0
    base, suffix = None, None

    # Loop through the files to find the highest version number
    for file in files:
        filename = os.path.basename(file)
        match = version_pattern.search(filename)
        if match:
            base, version_str, suffix = match.groups()
            version_num = int(version_str)
            if version_num > highest_version:
                highest_version = version_num

    # If no versioned files were found, use the provided filename and extension
    if base is None or suffix is None:
        base = f"{fname}{version_prefix}"
        suffix = ext
        highest_version = 0  # No previous versions

    # Increment the highest version number
    next_version_number = highest_version + 1

    # Format the next version number with the same number of digits as the original
    next_version_str = f"{base}{next_version_number:03d}{suffix}"

    # Combine the directory and new filename to create the full path
    next_filepath = os.path.join(dirname, next_version_str)

    return next_filepath


if __name__ == "__main__":
    # Start
    CONFIG, sys.stdout, sys.stderr, plt, CC = mngs.gen.start(sys, plt)

    # (YOUR AWESOME CODE)

    # Close
    mngs.gen.close(CONFIG)

# EOF

"""
/ssh:ywatanabe@444:/home/ywatanabe/proj/entrance/mngs/path/_version.py
"""



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Time-stamp: "2024-10-19 04:23:40 (ywatanabe)"
# /home/ywatanabe/proj/_mngs_repo_openhands/src/mngs/general/system_ops/_email.py

import os
import smtplib
from email import encoders
from email.mime.base import MIMEBase
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
import mimetypes
import mngs
import re

ansi_escape = re.compile(r'\x1B(?:[@-Z\\-_]|\[[0-?]*[ -/]*[@-~])')


def send_gmail(
    sender_gmail,
    sender_password,
    recipient_email,
    subject,
    message,
    sender_name=None,
    cc=None,
    ID=None,
    attachment_paths=None,
    verbose=True,
):
    if ID == "auto":
        ID = mngs.gen.gen_ID()

    if ID:
        if subject:
            subject = f"{subject} (ID: {ID})"
        else:
            subject = f"ID: {ID}"

    try:
        server = smtplib.SMTP("smtp.gmail.com", 587)
        server.starttls()
        server.login(sender_gmail, sender_password)

        gmail = MIMEMultipart()
        gmail["Subject"] = subject
        gmail["To"] = recipient_email
        if cc:
            if isinstance(cc, str):
                gmail["Cc"] = cc
            elif isinstance(cc, list):
                gmail["Cc"] = ", ".join(cc)
        if sender_name:
            gmail["From"] = f"{sender_name} <{sender_gmail}>"
        else:
            gmail["From"] = sender_gmail
        gmail_body = MIMEText(message, "plain")
        gmail.attach(gmail_body)


        # Attachment files
        if attachment_paths:
            for path in attachment_paths:
                _, ext = os.path.splitext(path)
                if ext.lower() == '.log':
                    with open(path, 'r', encoding='utf-8') as file:
                        content = file.read()
                        cleaned_content = ansi_escape.sub('', content)
                        part = MIMEText(cleaned_content, 'plain')

                        # part = MIMEText(file.read(), 'plain')
                else:
                    mime_type, _ = mimetypes.guess_type(path)
                    if mime_type is None:
                        mime_type = 'text/plain'
                    main_type, sub_type = mime_type.split('/', 1)
                    with open(path, "rb") as file:
                        part = MIMEBase(main_type, sub_type)
                        part.set_payload(file.read())
                        encoders.encode_base64(part)

                part.add_header(
                    "Content-Disposition",
                    f"attachment; filename={os.path.basename(path)}",
                )
                gmail.attach(part)

        recipients = [recipient_email]
        if cc:
            if isinstance(cc, str):
                recipients.append(cc)
            elif isinstance(cc, list):
                recipients.extend(cc)
        server.send_message(gmail, to_addrs=recipients)

        server.quit()

        if verbose:
            cc_info = f" (CC: {cc})" if cc else ""
            message = f"Email was sent:\n"
            message += f"    {sender_gmail} -> {recipient_email}{cc_info}\n"
            message += f"    (ID: {ID})\n"
            if attachment_paths:
                message += f"    Attached:\n"
                for ap in attachment_paths:
                    message += f"        {ap}\n"
            print(message)

            # message = f"\nEmail was sent:\n\t{sender_gmail} -> {recipient_email}{cc_info}\n\t(ID: {ID})"
            # if attachment_paths:
            #     attachment_paths_str = '\n\t\t'.join(attachment_paths)
            #     message += f"\n\tAttached:\n\t{attachment_paths_str}"
            # print(message)

    except Exception as e:
        print(f"Email was not sent: {e}")


# EOF



#!./env/bin/python3
# -*- coding: utf-8 -*-
# Time-stamp: "2024-07-29 21:43:30 (ywatanabe)"
# /home/ywatanabe/proj/mngs_repo/src/mngs/web/_crawl.py


import requests
from bs4 import BeautifulSoup
import urllib.parse
from concurrent.futures import ThreadPoolExecutor, as_completed
import json
from tqdm import tqdm
import mngs
from pprint import pprint
from readability import Document

import re


# def crawl_url(url, max_depth=1):
#     print("\nCrawling...")
#     visited = set()
#     to_visit = [(url, 0)]
#     contents = {}

#     while to_visit:
#         current_url, depth = to_visit.pop(0)
#         if current_url in visited or depth > max_depth:
#             continue

#         try:
#             response = requests.get(current_url)
#             if response.status_code == 200:
#                 visited.add(current_url)
#                 contents[current_url] = response.text
#                 soup = BeautifulSoup(response.text, "html.parser")

#                 for link in soup.find_all("a", href=True):
#                     absolute_link = urllib.parse.urljoin(
#                         current_url, link["href"]
#                     )
#                     if absolute_link not in visited:
#                         to_visit.append((absolute_link, depth + 1))

#         except requests.RequestException:
#             pass

#     return visited, contents


def extract_main_content(html):
    doc = Document(html)
    content = doc.summary()
    # Remove HTML tags
    content = re.sub("<[^<]+?>", "", content)
    # Remove extra whitespace
    content = " ".join(content.split())
    return content


def crawl_url(url, max_depth=1):
    print("\nCrawling...")
    visited = set()
    to_visit = [(url, 0)]
    contents = {}

    while to_visit:
        current_url, depth = to_visit.pop(0)
        if current_url in visited or depth > max_depth:
            continue

        try:
            response = requests.get(current_url)
            if response.status_code == 200:
                visited.add(current_url)
                main_content = extract_main_content(response.text)
                contents[current_url] = main_content
                soup = BeautifulSoup(response.text, "html.parser")

                for link in soup.find_all("a", href=True):
                    absolute_link = urllib.parse.urljoin(
                        current_url, link["href"]
                    )
                    if absolute_link not in visited:
                        to_visit.append((absolute_link, depth + 1))

        except requests.RequestException:
            pass

    return visited, contents


def crawl_to_json(start_url):
    if not start_url.startswith("http"):
        start_url = "https://" + start_url
    crawled_urls, contents = crawl_url(start_url)

    print("\nSummalizing as json...")

    def process_url(url):
        llm = mngs.ai.GenAI("gpt-4o-mini")
        return {
            "url": url,
            "content": llm(
                f"Summarize this page in 1 line:\n\n{contents[url]}"
            ),
        }

    with ThreadPoolExecutor() as executor:
        future_to_url = {
            executor.submit(process_url, url): url for url in crawled_urls
        }
        crawled_pages = []
        for future in tqdm(
            as_completed(future_to_url),
            total=len(crawled_urls),
            desc="Processing URLs",
        ):
            crawled_pages.append(future.result())

    result = {"start_url": start_url, "crawled_pages": crawled_pages}

    return json.dumps(result, indent=2)


def summarize_all(json_contents):
    llm = mngs.ai.GenAI("gpt-4o-mini")
    out = llm(
        f"Summarize this json file with 5 bullet points:\n\n{json_contents}"
    )
    return out


def summarize_url(start_url):
    json_result = crawl_to_json(start_url)
    ground_summary = summarize_all(json_result)

    pprint(ground_summary)
    return ground_summary, json_result


main = summarize_url

if __name__ == "__main__":
    import argparse
    import mngs

    parser = argparse.ArgumentParser(description="")
    parser.add_argument("--url", "-u", type=str, help="(default: %(default)s)")
    args = parser.parse_args()
    mngs.gen.print_block(args, c="yellow")

    main(args.url)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Time-stamp: "2024-10-07 22:30:46 (ywatanabe)"
# File: /home/ywatanabe/proj/_mngs_repo_openhands/src/mngs/stats/_brunner_munzel_test.py

"""
1. Functionality:
   - Calculates Brunner-Munzel test scores for two independent samples
2. Input:
   - Two arrays of numeric data values
3. Output:
   - Dictionary containing test results (w_statistic, p_value, sample sizes, degrees of freedom, effect size, test name, and null hypothesis)
4. Prerequisites:
   - NumPy, SciPy
"""

"""Imports"""
import numpy as np
from scipy import stats
import pandas as pd
import xarray as xr
import torch
from mngs.typing import List, Tuple, Dict, Any, Union, Sequence, Literal, Iterable, ArrayLike
from mngs.gen import numpy_fn

@numpy_fn
def brunner_munzel_test(
    x1: ArrayLike,
    x2: ArrayLike,
    distribution: str = "t",
    round_factor: int = 3,
) -> Dict[str, Union[float, int, str]]:
    """
    Calculate Brunner-Munzel test scores.

    Parameters
    ----------
    x1 : ArrayLike
        Numeric data values from sample 1.
    x2 : ArrayLike
        Numeric data values from sample 2.
    distribution : str, optional
        Distribution to use for the test. Can be "t" or "normal" (default is "t").
    round_factor : int, optional
        Number of decimal places to round the results (default is 3).

    Returns
    -------
    Dict[str, Union[float, int, str]]
        Dictionary containing test results including w_statistic, p_value, sample sizes, degrees of freedom, effect size, test name, and null hypothesis.

    Example
    -------
    >>> np.random.seed(42)
    >>> xx = np.random.rand(100)
    >>> yy = np.random.rand(100) + 0.1
    >>> result = brunner_munzel_test(xx, yy)
    >>> print(result)
    {'w_statistic': -2.089, 'p_value': 0.038, 'n1': 100, 'n2': 100, 'dof': 197.0, 'effsize': 0.438, 'test_name': 'Brunner-Munzel test', 'H0': 'The probability that a randomly selected value from one population is greater than a randomly selected value from the other population is equal to 0.5'}
    """
    if distribution not in ["t", "normal"]:
        raise ValueError("Distribution must be either 't' or 'normal'")

    x1, x2 = np.asarray(x1).astype(float), np.asarray(x2).astype(float)
    x1, x2 = x1[~np.isnan(x1)], x2[~np.isnan(x2)]
    n1, n2 = len(x1), len(x2)

    if n1 == 0 or n2 == 0:
        raise ValueError(
            "Input arrays must not be empty after removing NaN values"
        )

    R = stats.rankdata(np.concatenate([x1, x2]))
    R1, R2 = R[:n1], R[n1:]
    r1_mean, r2_mean = np.mean(R1), np.mean(R2)
    Ri1, Ri2 = stats.rankdata(x1), stats.rankdata(x2)
    var1 = np.var(R1 - Ri1, ddof=1)
    var2 = np.var(R2 - Ri2, ddof=1)

    w_statistic = ((n1 * n2) * (r2_mean - r1_mean)) / (
        (n1 + n2) * np.sqrt(n1 * var1 + n2 * var2)
    )

    if distribution == "t":
        dof = (n1 * var1 + n2 * var2) ** 2 / (
            (n1 * var1) ** 2 / (n1 - 1) + (n2 * var2) ** 2 / (n2 - 1)
        )
        c = (
            stats.t.cdf(abs(w_statistic), dof)
            if not np.isinf(w_statistic)
            else 0.0
        )
    else:
        dof = np.nan
        c = (
            stats.norm.cdf(abs(w_statistic))
            if not np.isinf(w_statistic)
            else 0.0
        )

    p_value = min(c, 1.0 - c) * 2.0
    effsize = (r2_mean - r1_mean) / (n1 + n2) + 0.5

    return {
        "w_statistic": round(w_statistic, round_factor),
        "p_value": round(p_value, round_factor),
        "n1": n1,
        "n2": n2,
        "dof": round(dof, round_factor),
        "effsize": round(effsize, round_factor),
        "test_name": "Brunner-Munzel test",
        "H0": "The probability that a randomly selected value from one population is greater than a randomly selected value from the other population is equal to 0.5",
    }


# #!/usr/bin/env python3

# import numpy as np
# from scipy import stats


# def brunner_munzel_test(x1, x2, distribution="t", round_factor=3):
#     """Calculate Brunner-Munzel-test scores.
#     Parameters:
#       x1, x2: array_like
#         Numeric data values from sample 1, 2.
#     Returns:
#       w:
#         Calculated test statistic.
#       p_value:
#         Two-tailed p-value of test.
#       dof:
#         Degree of freedom.
#       p:
#         "P(x1 < x2) + 0.5 P(x1 = x2)" estimates.
#     References:
#       * https://oku.edu.mie-u.ac.jp/~okumura/stat/brunner-munzel.html
#     Example:
#       When sample number N is small, distribution='t' is recommended.
#       d1 = np.array([1,2,1,1,1,1,1,1,1,1,2,4,1,1])
#       d2 = np.array([3,3,4,3,1,2,3,1,1,5,4])
#       print(bmtest(d1, d2, distribution='t'))
#       print(bmtest(d1, d2, distribution='normal'))
#       When sample number N is large, distribution='normal' is recommended; however,
#       't' and 'normal' yield almost the same result.
#       d1 = np.random.rand(1000)*100
#       d2 = np.random.rand(10000)*110
#       print(bmtest(d1, d2, distribution='t'))
#       print(bmtest(d1, d2, distribution='normal'))
#     """

#     x1, x2 = np.hstack(x1), np.hstack(x2)
#     x1, x2 = x1[~np.isnan(x1)], x2[~np.isnan(x2)]
#     n1, n2 = len(x1), len(x2)
#     R = stats.rankdata(list(x1) + list(x2))
#     R1, R2 = R[:n1], R[n1:]
#     r1_mean, r2_mean = np.mean(R1), np.mean(R2)
#     Ri1, Ri2 = stats.rankdata(x1), stats.rankdata(x2)
#     var1 = np.var([r - ri for r, ri in zip(R1, Ri1)], ddof=1)
#     var2 = np.var([r - ri for r, ri in zip(R2, Ri2)], ddof=1)
#     w_statistic = ((n1 * n2) * (r2_mean - r1_mean)) / (
#         (n1 + n2) * np.sqrt(n1 * var1 + n2 * var2)
#     )
#     if distribution == "t":
#         dof = (n1 * var1 + n2 * var2) ** 2 / (
#             (n1 * var1) ** 2 / (n1 - 1) + (n2 * var2) ** 2 / (n2 - 1)
#         )
#         c = (
#             stats.t.cdf(abs(w_statistic), dof)
#             if not np.isinf(w_statistic)
#             else 0.0
#         )
#     if distribution == "normal":
#         dof = np.nan
#         c = (
#             stats.norm.cdf(abs(w_statistic))
#             if not np.isinf(w_statistic)
#             else 0.0
#         )
#     p_value = min(c, 1.0 - c) * 2.0
#     effsize = (r2_mean - r1_mean) / (n1 + n2) + 0.5
#     return dict(
#         w_statistic=round(w_statistic, round_factor),
#         p_value=round(p_value, round_factor),
#         n1=n1,
#         n2=n2,
#         dof=round(dof, round_factor),
#         effsize=round(effsize, round_factor),
#         test_name="Brunner-Munzel test",
#         H0="The probability that a randomly selected value from one population is greater than a randomly selected value from the other population is equal to 0.5",
#     )



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Time-stamp: "2024-09-06 23:33:56 (ywatanabe)"

import matplotlib.pyplot as plt
import mngs
import numpy as np


# fixme, quite slow
def set_ticks(ax, xvals=None, xticks=None, yvals=None, yticks=None):
    ax = set_x_ticks(ax, x_vals=xvals, x_ticks=xticks)
    ax = set_y_ticks(ax, y_vals=yvals, y_ticks=yticks)
    canvas_type = type(ax.figure.canvas).__name__
    if "TkAgg" in canvas_type:
        ax.get_figure().canvas.draw()  # Redraw the canvas once after making all updates
    return ax


def set_x_ticks(ax, x_vals=None, x_ticks=None):
    """
    Set custom tick labels on the x and y axes based on specified values and desired ticks.

    Parameters:
    - ax: The axis object to modify.
    - x_vals: Array of x-axis values.
    - x_ticks: List of desired tick labels on the x-axis.
    - y_vals: Array of y-axis values.
    - y_ticks: List of desired tick labels on the y-axis.

    Example:
        import matplotlib.pyplot as plt
        import numpy as np

        fig, axes = plt.subplots(nrows=4)
        x = np.linspace(0, 10, 100)
        y = np.sin(x)
        for ax in axes:
            ax.plot(x, y)  # Plot a sine wave

        set_ticks(axes[0])  # Do nothing # OK
        set_ticks(axes[1], x_vals=x+3) # OK
        set_ticks(axes[2], x_ticks=[1,2])  # OK
        set_ticks(axes[3], x_vals=x+3, x_ticks=[4,5])  # Auto-generate ticks across the range
        fig.tight_layout()
        plt.show()
    """

    def _avoid_overlaps(values):
        values = np.array(values)
        if ("int" in str(values.dtype)) or ("float" in str(values.dtype)):
            values = values.astype(float) + np.arange(len(values)) * 1e-5
        return values

    def _set_x_vals(ax, x_vals):
        x_vals = _avoid_overlaps(x_vals)
        new_x_axis = np.linspace(*ax.get_xlim(), len(x_vals))
        ax.set_xticks(new_x_axis)
        ax.set_xticklabels([f"{xv}" for xv in x_vals])
        return ax

    def _set_x_ticks(ax, x_ticks):
        x_ticks = np.array(x_ticks)
        if x_ticks.dtype.kind in ["U", "S", "O"]:  # If x_ticks are strings
            ax.set_xticks(range(len(x_ticks)))
            ax.set_xticklabels(x_ticks)
        else:
            x_vals = np.array(
                [
                    label.get_text().replace("−", "-")
                    for label in ax.get_xticklabels()
                ]
            )
            x_vals = x_vals.astype(float)
            x_indi = np.argmin(
                np.array(
                    np.abs(x_vals[:, np.newaxis] - x_ticks[np.newaxis, :])
                ),
                axis=0,
            )
            ax.set_xticks(ax.get_xticks()[x_indi])
            ax.set_xticklabels([f"{xt}" for xt in x_ticks])
        return ax

    x_vals_passed = x_vals is not None
    x_ticks_passed = x_ticks is not None

    if mngs.gen.is_listed_X(x_ticks, dict):
        x_ticks = [mngs.gen.dict2str(xt, delimiter="\n") for xt in x_ticks]

    if (not x_vals_passed) and (not x_ticks_passed):
        # Do nothing
        pass

    elif x_vals_passed and (not x_ticks_passed):
        # Replaces the x axis to x_vals
        x_ticks = np.linspace(x_vals[0], x_vals[-1], 4)
        ax = _set_x_vals(ax, x_ticks)

    elif (not x_vals_passed) and x_ticks_passed:
        # Locates 'x_ticks' on the original x axis
        ax.set_xticks(x_ticks)

    elif x_vals_passed and x_ticks_passed:
        if x_vals == "auto":
            x_vals = np.arange(len(x_ticks))

        # Replaces the original x axis to 'x_vals' and locates the 'x_ticks' on the new axis
        ax = _set_x_vals(ax, x_vals)
        ax = _set_x_ticks(ax, x_ticks)

    return ax


def set_y_ticks(ax, y_vals=None, y_ticks=None):
    """
    Set custom tick labels on the y-axis based on specified values and desired ticks.

    Parameters:
    - ax: The axis object to modify.
    - y_vals: Array of y-axis values where ticks should be placed.
    - y_ticks: List of labels for ticks on the y-axis.

    Example:
        import matplotlib.pyplot as plt
        import numpy as np

        fig, ax = plt.subplots()
        x = np.linspace(0, 10, 100)
        y = np.sin(x)
        ax.plot(x, y)  # Plot a sine wave

        set_y_ticks(ax, y_vals=y, y_ticks=['Low', 'High'])  # Set custom y-axis ticks
        plt.show()
    """

    def _avoid_overlaps(values):
        values = np.array(values)
        if ("int" in str(values.dtype)) or ("float" in str(values.dtype)):
            values = values.astype(float) + np.arange(len(values)) * 1e-5
        return values

    def _set_y_vals(ax, y_vals):
        y_vals = _avoid_overlaps(y_vals)
        new_y_axis = np.linspace(*ax.get_ylim(), len(y_vals))
        ax.set_yticks(new_y_axis)
        ax.set_yticklabels([f"{yv:.2f}" for yv in y_vals])
        return ax

    # def _set_y_ticks(ax, y_ticks):
    #     y_ticks = np.array(y_ticks)
    #     y_vals = np.array(
    #         [
    #             label.get_text().replace("−", "-")
    #             for label in ax.get_yticklabels()
    #         ]
    #     )
    #     y_vals = y_vals.astype(float)
    #     y_indi = np.argmin(
    #         np.array(np.abs(y_vals[:, np.newaxis] - y_ticks[np.newaxis, :])),
    #         axis=0,
    #     )

    #     # y_indi = [np.argmin(np.abs(y_vals - yt)) for yt in y_ticks]
    #     ax.set_yticks(ax.get_yticks()[y_indi])
    #     ax.set_yticklabels([f"{yt}" for yt in y_ticks])
    #     return ax
    def _set_y_ticks(ax, y_ticks):
        y_ticks = np.array(y_ticks)
        if y_ticks.dtype.kind in ["U", "S", "O"]:  # If y_ticks are strings
            ax.set_yticks(range(len(y_ticks)))
            ax.set_yticklabels(y_ticks)
        else:
            y_vals = np.array(
                [
                    label.get_text().replace("−", "-")
                    for label in ax.get_yticklabels()
                ]
            )
            y_vals = y_vals.astype(float)
            y_indi = np.argmin(
                np.array(
                    np.abs(y_vals[:, np.newaxis] - y_ticks[np.newaxis, :])
                ),
                axis=0,
            )
            ax.set_yticks(ax.get_yticks()[y_indi])
            ax.set_yticklabels([f"{yt}" for yt in y_ticks])
        return ax

    y_vals_passed = y_vals is not None
    y_ticks_passed = y_ticks is not None

    if mngs.gen.is_listed_X(y_ticks, dict):
        y_ticks = [mngs.gen.dict2str(yt, delimiter="\n") for yt in y_ticks]

    if (not y_vals_passed) and (not y_ticks_passed):
        # Do nothing
        pass

    elif y_vals_passed and (not y_ticks_passed):
        # Replaces the y axis to y_vals
        ax = _set_y_vals(ax, y_vals)

    elif (not y_vals_passed) and y_ticks_passed:
        # Locates 'y_ticks' on the original y axis
        ax.set_yticks(y_ticks)

    elif y_vals_passed and y_ticks_passed:
        # Replaces the original y axis to 'y_vals' and locates the 'y_ticks' on the new axis
        if y_vals == "auto":
            y_vals = np.arange(len(y_ticks))

        ax = _set_y_vals(ax, y_vals)
        ax = _set_y_ticks(ax, y_ticks)
    return ax


if __name__ == "__main__":
    import mngs

    xx, tt, fs = mngs.dsp.demo_sig()
    pha, amp, freqs = mngs.dsp.wavelet(xx, fs)

    i_batch, i_ch = 0, 0
    ff = freqs[i_batch, i_ch]
    fig, ax = mngs.plt.subplots()

    ax.imshow2d(amp[i_batch, i_ch])

    ax = set_ticks(
        ax,
        x_vals=tt,
        x_ticks=[0, 1, 2, 3, 4],
        y_vals=ff,
        y_ticks=[0, 128, 256],
    )

    plt.show()



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Time-stamp: "2024-10-13 19:04:49 (ywatanabe)"
# File: 2024-10-02 23:28:44

import os
import importlib
import inspect

# Get the current directory
current_dir = os.path.dirname(__file__)

# Iterate through all Python files in the current directory
for filename in os.listdir(current_dir):
    if filename.endswith(".py") and not filename.startswith("__"):
        module_name = filename[:-3]  # Remove .py extension
        # module = importlib.import_module(f".{module_name}", package=__name__)
        # # Import only functions and classes from the module
        # for name, obj in inspect.getmembers(module):
        #     if inspect.isfunction(obj) or inspect.isclass(obj):
        #         if not name.startswith("_"):
        #             globals()[name] = obj

        try:
            module = importlib.import_module(f".{module_name}", package=__name__)

            # Import only functions and classes from the module
            for name, obj in inspect.getmembers(module):
                if inspect.isfunction(obj) or inspect.isclass(obj):
                    if not name.startswith("_"):
                        globals()[name] = obj
        except ImportError as e:
            warnings.warn(f"Warning: Failed to import {module_name}. Error: {e}")

# Clean up temporary variables
del os, importlib, inspect, current_dir, filename, module_name, module, name, obj

# EOF



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Time-stamp: "2024-09-29 05:55:52 (ywatanabe)"
# /home/ywatanabe/proj/_mngs_repo_openhands/src/mngs/pd/_force_df.py

import numpy as np
import pandas as pd
import mngs

# def force_df(permutable_dict, filler=""):
def force_df(permutable_dict, filler=np.nan):

    if mngs.gen.is_listed_X(permutable_dict, pd.Series):
        permutable_dict = [sr.to_dict() for sr in permutable_dict]
    ## Deep copy
    permutable_dict = permutable_dict.copy()

    ## Get the lengths
    max_len = 0
    for k, v in permutable_dict.items():
        # Check if v is an iterable (but not string) or treat as single length otherwise
        if isinstance(v, (str, int, float)) or not hasattr(v, "__len__"):
            length = 1
        else:
            length = len(v)
        max_len = max(max_len, length)

    ## Replace with appropriately filled list
    for k, v in permutable_dict.items():
        if isinstance(v, (str, int, float)) or not hasattr(v, "__len__"):
            permutable_dict[k] = [v] + [filler] * (max_len - 1)
        else:
            permutable_dict[k] = list(v) + [filler] * (max_len - len(v))

    ## Puts them into a DataFrame
    out_df = pd.DataFrame(permutable_dict)

    return out_df



import os
from contextlib import contextmanager, redirect_stderr, redirect_stdout

@contextmanager
def suppress_output(suppress=True):
    """
    A context manager that suppresses stdout and stderr.

    Example:
        with suppress_output():
            print("This will not be printed to the console.")
    """
    if suppress:
        # Open a file descriptor that points to os.devnull (a black hole for data)
        with open(os.devnull, "w") as fnull:
            # Temporarily redirect stdout and stderr to the file descriptor fnull
            with redirect_stdout(fnull), redirect_stderr(fnull):
                # Yield control back to the context block
                yield
    else:
        # If suppress is False, just yield without redirecting output
        yield

quiet = suppress_output



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Time-stamp: "2024-10-06 09:38:04 (ywatanabe)"
# /home/ywatanabe/proj/_mngs_repo_openhands/src/mngs/stats/_find_pval_col.py

"""
Functionality:
    - Identifies the column name in a DataFrame that corresponds to p-values
Input:
    - pandas DataFrame, array-like object, list, or dict
Output:
    - String representing the identified p-value column name, or None if not found
Prerequisites:
    - pandas library
"""

import re
from typing import Optional, Union, List, Dict
import pandas as pd
import numpy as np

def find_pval(data: Union[pd.DataFrame, np.ndarray, List, Dict]) -> Optional[str]:
    """
    Find the column name or key that matches p-value patterns.

    Example:
    --------
    >>> import pandas as pd
    >>> df = pd.DataFrame({'p_value': [0.05, 0.01], 'pval': [0.1, 0.001], 'other': [1, 2]})
    >>> find_pval(df)
    'p_value'

    Parameters:
    -----------
    data : Union[pd.DataFrame, np.ndarray, List, Dict]
        Data structure to search for p-value column or key

    Returns:
    --------
    Optional[str]
        Name of the column or key that matches p-value patterns, or None if not found
    """
    if isinstance(data, pd.DataFrame):
        return find_pval_col(data)
    elif isinstance(data, (np.ndarray, list, dict)):
        return _find_pval(data)
    else:
        raise ValueError("Input must be a pandas DataFrame, numpy array, list, or dict")

def _find_pval(data: Union[np.ndarray, List, Dict]) -> Optional[str]:
    pattern = re.compile(r'p[-_]?val(ue)?', re.IGNORECASE)
    if isinstance(data, dict):
        for key in data.keys():
            if pattern.search(str(key)):
                return key
    elif isinstance(data, (np.ndarray, list)):
        if len(data) > 0 and isinstance(data[0], dict):
            for key in data[0].keys():
                if pattern.search(str(key)):
                    return key
    return None

def find_pval_col(df: pd.DataFrame) -> Optional[str]:
    """
    Find the column name that matches p-value patterns in a DataFrame.

    Example:
    --------
    >>> import pandas as pd
    >>> df = pd.DataFrame({'p_value': [0.05, 0.01], 'pval': [0.1, 0.001], 'other': [1, 2]})
    >>> find_pval_col(df)
    'p_value'

    Parameters:
    -----------
    df : pd.DataFrame
        DataFrame to search for p-value column

    Returns:
    --------
    Optional[str]
        Name of the column that matches p-value patterns, or None if not found
    """
    pattern = re.compile(r'p[-_]?val(ue)?', re.IGNORECASE)
    for col in df.columns:
        if pattern.search(str(col)):
            return col
    return None



#!/usr/bin/env python3


import matplotlib as mpl
import matplotlib.pyplot as plt
from matplotlib import cm


class ColorGetter:
    # https://stackoverflow.com/questions/26108436/how-can-i-get-the-matplotlib-rgb-color-given-the-colormap-name-boundrynorm-an
    def __init__(self, cmap_name, start_val, stop_val):
        self.cmap_name = cmap_name
        self.cmap = plt.get_cmap(cmap_name)
        self.norm = mpl.colors.Normalize(vmin=start_val, vmax=stop_val)
        self.scalarMap = cm.ScalarMappable(norm=self.norm, cmap=self.cmap)

    def get_rgb(self, val):
        return self.scalarMap.to_rgba(val)


def get_RGBA_from_colormap(val, cmap="Blues", cmap_start_val=0, cmap_stop_val=1):
    ColGetter = ColorGetter(cmap, cmap_start_val, cmap_stop_val)
    return ColGetter.get_rgb(val)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Time-stamp: "2024-09-10 20:05:21 (ywatanabe)"

import inspect
import os
import pwd
import smtplib
import socket
import sys
from email import encoders
from email.mime.base import MIMEBase
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText

import mngs


def get_username():
    try:
        return pwd.getpwuid(os.getuid()).pw_name
    except:
        return os.getenv("USER") or os.getenv("LOGNAME") or "unknown"


def get_hostname():
    return socket.gethostname()


def get_git_branch():
    try:
        return (
            subprocess.check_output(
                ["git", "rev-parse", "--abbrev-ref", "HEAD"],
                cwd=mngs.__path__[0],
                stderr=subprocess.DEVNULL,
            )
            .decode()
            .strip()
        )
    except:
        return "main"


def notify(
    subject="",
    message=":)",
    ID="auto",
    recipient_name=None,
    recipient_email=None,
    cc=None,
    log_paths=None,
    verbose=False,
):
    """
    Usage:
        notify(subject="subject here", message="message here", recipient_email="example@email.com", cc=["cc1@email.com", "cc2@email.com"])

    Note:
        This function operates correctly when a Gmail address is configured as follows:

        Step 1: Set up 2-Step Verification (if not already active):
            1. Navigate to your Google Account.
            2. Click on "Security".
            3. Under "Signing in to Google", select "2-Step Verification".
            4. Follow the on-screen instructions.

        Step 2: Generate an App Password
            1. Return to the "Security" section of your Google Account.
            2. Under "Signing in to Google", choose "App Passwords". You may need to sign in again.
            3. At the bottom, click "Select app" and pick the app you are using (select "Mail" for general purposes).
            4. Click "Select device" and select the device you are using (or choose "Other" and label it "Python Script" or similar).
            5. Click "Generate".
            6. Use the 16-digit App Password in your script as directed. The App Password will be verbosen only once, so remember to copy it.

        Step 3: Configure Gmail as environment variables
               ```bash
               export MNGS_SENDER_GMAIL="mngs.notification@gmail.com"
               export MNGS_SENDER_GMAIL_PASSWORD="YOUR_APP_PASSWORD"
               export MNGS_RECIPIENT_GMAIL="YOUR_GMAIL_ADDRESS"
               ```
    """
    # Environmental variables
    sender_gmail = os.getenv("MNGS_SENDER_GMAIL")
    sender_password = os.getenv("MNGS_SENDER_GMAIL_PASSWORD")
    recipient_email = recipient_email or os.getenv("MNGS_RECIPIENT_GMAIL")

    # Get the script name from sys.argv or inspect the stack
    if sys.argv[0]:
        script_name = os.path.basename(sys.argv[0])
    else:
        frames = inspect.stack()
        script_name = (
            os.path.basename(frames[-1].filename) if frames else "(Not found)"
        )
    if (script_name == "-c") or (script_name.endswith(".py")):
        script_name = "`$ python -c ...`"

    # Recipient name
    recipient_name = recipient_name if recipient_name is not None else "there"

    sender = f"{get_username()}@{get_hostname()}"
    header = f"Hi {recipient_name} 👋\n\n"
    branch = get_git_branch()
    footer = f"""

Best regards,
{sender}

{'-'*20}
Sent via
- Script: {script_name}
- Source: mngs v{mngs.__version__} (github.com/ywatanabe1989/mngs/blob/{branch}/src/mngs/general/_email.py)
{'-'*20}"""

    full_message = header + message + footer
    full_subject = f"{subject}"

    if sender_gmail is None or sender_password is None:
        print(
            f"""
        Please set environmental variables to use this function (f"{inspect.stack()[0][3]}"):\n\n
        $ export MNGS_SENDER_GMAIL="mngs.notification@gmail.com"
        $ export MNGS_SENDER_GMAIL_PASSWORD="YOUR_APP_PASSWORD"
        $ export MNGS_RECIPIENT_GMAIL="YOUR_GMAIL_ADDRESS"
        """
        )

    send_gmail(
        sender_gmail,
        sender_password,
        recipient_email,
        full_subject,
        full_message,
        cc=cc,
        ID=ID,
        log_paths=log_paths,
        verbose=verbose,
    )


def send_gmail(
    sender_gmail,
    sender_password,
    recipient_email,
    subject,
    message,
    ID=None,
    cc=None,
    log_paths=None,
    verbose=True,
):

    if ID == "auto":
        ID = mngs.gen.gen_ID()  # .split("_")[-1]
        subject = f"{subject} (ID: {ID})"

    try:
        # Set up the gmail server
        server = smtplib.SMTP("smtp.gmail.com", 587)
        server.starttls()

        # Login to the gmail account
        server.login(sender_gmail, sender_password)

        # Create the gmail
        gmail = MIMEMultipart()
        gmail["From"] = sender_gmail
        gmail["To"] = recipient_email
        if cc:
            if isinstance(cc, str):
                gmail["Cc"] = cc
            elif isinstance(cc, list):
                gmail["Cc"] = ", ".join(cc)
        gmail["Subject"] = subject
        gmail_body = MIMEText(message, "plain")
        gmail.attach(gmail_body)

        # Attach log files if provided
        if log_paths:
            for path in log_paths:
                with open(path, "rb") as file:
                    part = MIMEBase("application", "octet-stream")
                    part.set_payload(file.read())
                    encoders.encode_base64(part)
                    part.add_header(
                        "Content-Disposition",
                        f"attachment; filename= {os.path.basename(path)}",
                    )
                    gmail.attach(part)

        # Send the email
        recipients = [recipient_email]
        if cc:
            if isinstance(cc, str):
                recipients.append(cc)
            elif isinstance(cc, list):
                recipients.extend(cc)
        print(cc)
        print(recipients)
        server.send_message(gmail, to_addrs=recipients)

        # Quit the server
        server.quit()

        # if verbose:
        #     print(
        #         f"\nEmail was sent:\n\t{sender_gmail} -> {recipient_gmail}\n\t(ID: {ID})"
        #     )
        if verbose:
            cc_info = f" (CC: {cc})" if cc else ""
            print(
                f"\nEmail was sent:\n\t{sender_gmail} -> {recipient_email}{cc_info}\n\t(ID: {ID})"
            )

    except Exception as e:
        print(f"Email was not sent: {e}")


if __name__ == "__main__":
    notify(verbose=True)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Time-stamp: "2023-04-23 09:45:28 (ywatanabe)"

import torch
import torch.nn as nn
import torch.nn.functional as F
from torchsummary import summary
import mngs
import numpy as np

class SpatialAttention(nn.Module):
    def __init__(
            self, n_chs_in
    ):
        super().__init__()
        self.aap = nn.AdaptiveAvgPool1d(1)
        self.conv11 = nn.Conv1d(in_channels=n_chs_in, out_channels=1, kernel_size=1)

    def forward(self, x):
        """x: [batch_size, n_chs, seq_len]"""
        x_orig = x
        x = self.aap(x)
        x = self.conv11(x)

        return x * x_orig



#!/usr/bin/env python3
# Time-stamp: "2024-10-27 13:08:58 (ywatanabe)"

########################################
# Warnings
########################################
import warnings
warnings.filterwarnings("ignore", message="There is no current event loop", category=DeprecationWarning)

########################################
# Try import
########################################
import os
from .gen._suppress_output import suppress_output
_do_suppress = os.getenv("MNGS_SUPPRESS_IMPORTING_MESSAGES", "").lower() == "true"
# with suppress_output(suppress=__do_suppress):

########################################
# Core Modules
########################################
from ._sh import sh
from . import io
from . import path
from . import general
from . import gen
from . import ai
from . import ml
from . import dsp
from . import gists
from . import linalg
from . import nn
from . import os
from . import plt
from . import stats
from . import torch
from . import tex
from . import typing
from . import res
from . import web
from . import db
from . import pd

# ########################################
# # Lazy importing
# ########################################
# from importlib import import_module
# class LazyModule:
#     def __init__(self, name):
#         self._name = name
#         self._mod = None

#     def __getattr__(self, attr):
#         if self._mod is None:
#             self._mod = import_module(f'.{self._name}', 'mngs')
#         return getattr(self._mod, attr)

# io = LazyModule('io')
# path = LazyModule('path')
# general = LazyModule('general')
# gen = LazyModule('gen')
# ai = LazyModule('ai')
# ml = LazyModule('ml')
# dsp = LazyModule('dsp')
# gists = LazyModule('gists')
# linalg = LazyModule('linalg')
# nn = LazyModule('nn')
# os = LazyModule('os')
# plt = LazyModule('plt')
# stats = LazyModule('stats')
# torch = LazyModule('torch')
# tex = LazyModule('tex')
# typing = LazyModule('typing')
# res = LazyModule('res')
# web = LazyModule('web')
# db = LazyModule('db')
# pd = LazyModule('pd')

# additional_modules = [
#     "io",
#     "path",
#     "general",
#     "gen",
#     "ai",
#     "ml",
#     "dsp",
#     "gists",
#     "linalg",
#     "nn",
#     "os",
#     "plt",
#     "stats",
#     "torch",
#     "tex",
#     "typing",
#     "res",
#     "web",
#     "db",
#     "pd",
# ]

# for module_name in additional_modules:
#     module = try_import(f".{module_name}")
#     if module:
#         globals()[module_name] = module

# with suppress_output(suppress=suppress):
#     try:
#         from ._sh import sh
#     except ImportError as e:
#         warnings.warn(f"Warning: Failed to import some core modules. Error: {e}")

#     # Additional modules
#     additional_modules = [
#         "io",
#         "path",
#         "general",
#         "gen",
#         "ai",
#         "ml",
#         "dsp",
#         "gists",
#         "linalg",
#         "nn",
#         "os",
#         "plt",
#         "stats",
#         "torch",
#         "tex",
#         "typing",
#         "res",
#         "web",
#         "db",
#         "pd",
#     ]
#     for module in additional_modules:
#         try:
#             exec(f"from . import {module}")
#         except ImportError as e:
#             warnings.warn(f"Warning: Failed to import {module}. Error: {e}")


#     from ._sh import sh


# Modules
from .gen._print_config import print_config
# Usage: python -m mngs print_config

__copyright__ = "Copyright (C) 2024 Yusuke Watanabe"
__version__ = "1.8.1"
__license__ = "MIT"
__author__ = "ywatanabe1989"
__author_email__ = "ywatanabe@alumni.u-tokyo.ac.jp"
__url__ = "https://github.com/ywatanabe1989/mngs"



# RangerQH - @lessw2020 github
# Combines Quasi Hyperbolic momentum with Hinton Lookahead.

# https://arxiv.org/abs/1810.06801v4  (QH paper)
# #Lookahead paper --> MZhang,G Hinton  https://arxiv.org/abs/1907.08610




# Some portions = Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import torch
from torch.optim.optimizer import Optimizer

#from ..common import param_conv


class RangerQH(Optimizer):
    r"""Implements the QHAdam optimization algorithm `(Ma and Yarats, 2019)`_.
    Along with Hinton/Zhang Lookahead.
    Args:
        params (iterable):
            iterable of parameters to optimize or dicts defining parameter
            groups
        lr (float, optional): learning rate (:math:`\alpha` from the paper)
            (default: 1e-3)
        betas (Tuple[float, float], optional): coefficients used for computing
            running averages of the gradient and its square
            (default: (0.9, 0.999))
        nus (Tuple[float, float], optional): immediate discount factors used to
            estimate the gradient and its square
            (default: (1.0, 1.0))
        eps (float, optional): term added to the denominator to improve
            numerical stability
            (default: 1e-8)
        weight_decay (float, optional): weight decay (default: 0.0)
        decouple_weight_decay (bool, optional): whether to decouple the weight
            decay from the gradient-based optimization step
            (default: False)
    Example:
        >>> optimizer = qhoptim.pyt.QHAdam(
        ...     model.parameters(),
        ...     lr=3e-4, nus=(0.8, 1.0), betas=(0.99, 0.999))
        >>> optimizer.zero_grad()
        >>> loss_fn(model(input), target).backward()
        >>> optimizer.step()
    .. _`(Ma and Yarats, 2019)`: https://arxiv.org/abs/1810.06801
    """

    def __init__(
        self,
        params,
        lr=1e-3,
        betas=(0.9, 0.999),
        nus=(.7, 1.0),
        weight_decay=0.0,
        k=6,
        alpha=.5,
        decouple_weight_decay=False,
        eps=1e-8,
    ):
        if not 0.0 <= lr:
            raise ValueError("Invalid learning rate: {}".format(lr))
        if not 0.0 <= eps:
            raise ValueError("Invalid epsilon value: {}".format(eps))
        if not 0.0 <= betas[0] < 1.0:
            raise ValueError("Invalid beta parameter at index 0: {}".format(betas[0]))
        if not 0.0 <= betas[1] < 1.0:
            raise ValueError("Invalid beta parameter at index 1: {}".format(betas[1]))
        if weight_decay < 0.0:
            raise ValueError("Invalid weight_decay value: {}".format(weight_decay))

        defaults = {
            "lr": lr,
            "betas": betas,
            "nus": nus,
            "weight_decay": weight_decay,
            "decouple_weight_decay": decouple_weight_decay,
            "eps": eps,
        }
        super().__init__(params, defaults)
        
        #look ahead params
        self.alpha = alpha
        self.k = k         
        

    def step(self, closure=None):
        """Performs a single optimization step.
        Args:
            closure (callable, optional):
                A closure that reevaluates the model and returns the loss.
        """
        loss = None
        if closure is not None:
            loss = closure()

        for group in self.param_groups:
            lr = group["lr"]
            beta1, beta2 = group["betas"]
            nu1, nu2 = group["nus"]
            weight_decay = group["weight_decay"]
            decouple_weight_decay = group["decouple_weight_decay"]
            eps = group["eps"]

            for p in group["params"]:
                if p.grad is None:
                    continue

                d_p = p.grad.data
                if d_p.is_sparse:
                    raise RuntimeError("QHAdam does not support sparse gradients")

                

                if weight_decay != 0:
                    if decouple_weight_decay:
                        p.data.mul_(1 - lr * weight_decay)
                    else:
                        d_p.add_(weight_decay, p.data)

                d_p_sq = d_p.mul(d_p)
                
                #prep for saved param loading
                param_state = self.state[p]

                if len(param_state) == 0:
                    param_state["beta1_weight"] = 0.0
                    param_state["beta2_weight"] = 0.0
                    param_state['step'] = 0
                    param_state["exp_avg"] = torch.zeros_like(p.data)
                    param_state["exp_avg_sq"] = torch.zeros_like(p.data)
                    #look ahead weight storage now in state dict 
                    param_state['slow_buffer'] = torch.empty_like(p.data)
                    param_state['slow_buffer'].copy_(p.data)
                
                
                param_state['step'] += 1                

                param_state["beta1_weight"] = 1.0 + beta1 * param_state["beta1_weight"]
                param_state["beta2_weight"] = 1.0 + beta2 * param_state["beta2_weight"]

                beta1_weight = param_state["beta1_weight"]
                beta2_weight = param_state["beta2_weight"]
                exp_avg = param_state["exp_avg"]
                exp_avg_sq = param_state["exp_avg_sq"]

                beta1_adj = 1.0 - (1.0 / beta1_weight)
                beta2_adj = 1.0 - (1.0 / beta2_weight)
                exp_avg.mul_(beta1_adj).add_(1.0 - beta1_adj, d_p)
                exp_avg_sq.mul_(beta2_adj).add_(1.0 - beta2_adj, d_p_sq)

                avg_grad = exp_avg.mul(nu1)
                if nu1 != 1.0:
                    avg_grad.add_(1.0 - nu1, d_p)

                avg_grad_rms = exp_avg_sq.mul(nu2)
                if nu2 != 1.0:
                    avg_grad_rms.add_(1.0 - nu2, d_p_sq)
                avg_grad_rms.sqrt_()
                if eps != 0.0:
                    avg_grad_rms.add_(eps)

                p.data.addcdiv_(-lr, avg_grad, avg_grad_rms)
                
                #integrated look ahead...
                #we do it at the param level instead of group level
                if param_state['step'] % self.k ==0: #group['k'] == 0:
                    slow_p = param_state['slow_buffer'] #get access to slow param tensor
                    slow_p.add_(self.alpha, p.data - slow_p)  #(fast weights - slow weights) * alpha
                    p.data.copy_(slow_p)  #copy interpolated weights to RAdam param tensor
                

        return loss

    @classmethod
    def _params_to_dict(cls, params):
        return {"lr": params.alpha, "nus": (params.nu1, params.nu2), "betas": (params.beta1, params.beta2)}




#!./env/bin/python3
# -*- coding: utf-8 -*-
# Time-stamp: "2024-08-26 08:38:55 (ywatanabe)"
# /home/ywatanabe/proj/mngs_repo/src/mngs/plt/ax/_joyplot.py

import joypy
import mngs

# def plot_joy(df, cols_NT):
#     df_plot = df[cols_NT]
#     fig, axes = joypy.joyplot(
#         data=df_plot,
#         colormap=plt.cm.viridis,
#         title="Distribution of Ranked Data",
#         labels=cols_NT,
#         overlap=0.5,
#         orientation="vertical",
#     )
#     plt.xlabel("Variables")
#     plt.ylabel("Rank")
#     return fig


def joyplot(ax, data, **kwargs):
    fig, axes = joypy.joyplot(
        data=data,
        **kwargs,
    )

    if kwargs.get("orientation") == "vertical":
        xlabel = None
        ylabel = "Density"
    else:
        xlabel = "Density"
        ylabel = None

    ax = mngs.plt.ax.set_xyt(ax, xlabel, ylabel, "Joyplot")

    return ax



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Time-stamp: "2024-10-19 17:15:38 (ywatanabe)"
# File: 2024-10-15 00:35:40

from ._inspect import inspect, Inspector
from ._delete_duplicates import delete_duplicates



#!./env/bin/python3
# -*- coding: utf-8 -*-
# Time-stamp: "2024-07-29 23:39:18 (ywatanabe)"
# /home/ywatanabe/proj/mngs_repo/src/mngs/web/_google.py


"""
This script does XYZ.
"""


"""
Imports
"""
import os
import re
import sys

import matplotlib
import matplotlib.pyplot as plt
import seaborn as sns
import importlib

import mngs

import numpy as np
import pandas as pd
import torch
import torch.nn as nn
import torch.nn.functional as F
from icecream import ic
from natsort import natsorted
from glob import glob
from pprint import pprint
import warnings
import logging
from tqdm import tqdm
import xarray as xr

# sys.path = ["."] + sys.path
# from scripts import utils, load

"""
Warnings
"""
# warnings.simplefilter("ignore", UserWarning)


"""
Config
"""
# CONFIG = mngs.gen.load_configs()


"""
Functions & Classes
"""


def main2():
    import requests
    from bs4 import BeautifulSoup

    query = "Python programming"
    url = f"https://www.google.com/search?q={query}"

    # Send a GET request
    response = requests.get(url)

    # Check if the request was successful
    if response.status_code == 200:
        # Parse the HTML
        soup = BeautifulSoup(response.text, "html.parser")

        # links = soup.find_all("a")
        # # Extract the URLs from href attributes
        # urls = []
        # for link in links:
        #     href = link.get("href")  # Get the href attribute
        #     if href:  # If href is not None
        #         urls.append(href)

        # # Print the extracted URLs
        # for u in urls:
        #     print(u)

        # Find all search result links
        for g in soup.find_all("h3"):
            print(g.get_text())
    else:
        print(f"Failed to retrieve results: {response.status_code}")


def main():
    from googlesearch import search

    # Query
    query = "Python programming"

    # Perform the search
    for result in search(query, num_results=10):
        print(result)
    pass


if __name__ == "__main__":
    # # Argument Parser
    # import argparse
    # parser = argparse.ArgumentParser(description='')
    # parser.add_argument('--var', '-v', type=int, default=1, help='')
    # parser.add_argument('--flag', '-f', action='store_true', default=False, help='')
    # args = parser.parse_args()

    # Main
    CONFIG, sys.stdout, sys.stderr, plt, CC = mngs.gen.start(
        sys, plt, verbose=False
    )
    main()
    # main2()
    mngs.gen.close(CONFIG, verbose=False, notify=False)

# EOF



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Time-stamp: "2024-09-15 09:39:02 (ywatanabe)"
# /home/ywatanabe/proj/_mngs_repo_openhands/src/mngs/plt/ax/_format_label.py


def format_label(label):
    """
    Format label by capitalizing first letter and replacing underscores with spaces.
    """

    # if isinstance(label, str):
    #     # Replace underscores with spaces
    #     label = label.replace("_", " ")

    #     # Capitalize first letter of each word
    #     label = " ".join(word.capitalize() for word in label.split())

    #     # Special case for abbreviations (all caps)
    #     if label.isupper():
    #         return label

    return label



#!/./env/bin/python3
# -*- coding: utf-8 -*-
# Time-stamp: "2024-09-03 07:01:31 (ywatanabe)"
# /home/ywatanabe/proj/mngs_repo/src/mngs/pd/_to_xy.py

import mngs
import numpy as np
import pandas as pd


def to_xy(data_frame):
    """
    Convert a heatmap DataFrame into x, y, z format.

    Ensure the index and columns are the same, and if either exists, replace with that.

    Example
    -------
    data_frame = pd.DataFrame(...)  # Your DataFrame here
    out = to_xy(data_frame)
    print(out)

    Parameters
    ----------
    data_frame : pandas.DataFrame
        The input DataFrame to be converted.

    Returns
    -------
    pandas.DataFrame
        A DataFrame formatted with columns ['x', 'y', 'z']
    """
    assert data_frame.shape[0] == data_frame.shape[1]

    if not data_frame.index.equals(data_frame.columns):

        if (data_frame.index == np.array(range(len(data_frame.index)))).all():
            data_frame.columns = data_frame.index
        elif (
            data_frame.columns == np.array(range(len(data_frame.columns)))
        ).all():
            data_frame.index = data_frame.columns
        else:
            ValueError
        # else:
        #     ValueError "Either of index or columns has to be passed"

    formatted_data_frames = []

    for column in data_frame.columns:
        column_data_frame = data_frame[column]
        y_label = column_data_frame.name
        column_data_frame = pd.DataFrame(column_data_frame)
        column_data_frame["x"] = column_data_frame.index
        column_data_frame["y"] = y_label
        column_data_frame = column_data_frame.reset_index().drop(
            columns=["index"]
        )
        column_data_frame = column_data_frame.rename(columns={y_label: "z"})
        column_data_frame = mngs.pd.mv(column_data_frame, "z", -1)
        formatted_data_frames.append(column_data_frame)

    return pd.concat(formatted_data_frames, ignore_index=True)



#!./env/bin/python3
# -*- coding: utf-8 -*-
# Time-stamp: "2024-04-10 11:59:49 (ywatanabe)"


def ensure_even_len(x):
    if x.shape[-1] % 2 == 0:
        return x
    else:
        return x[..., :-1]



#!./env/bin/python3
# -*- coding: utf-8 -*-
# Time-stamp: "2024-07-10 10:33:54 (ywatanabe)"
# /home/ywatanabe/proj/mngs_repo/src/mngs/plt/_check_colors.py


"""
This script does XYZ.
"""


"""
Imports
"""
import os
import re
import sys

import matplotlib
import matplotlib.pyplot as plt
import seaborn as sns
import importlib

import mngs

importlib.reload(mngs)

import numpy as np
import pandas as pd
import torch
import torch.nn as nn
import torch.nn.functional as F
from icecream import ic
from natsort import natsorted
from glob import glob
from pprint import pprint
import warnings
import logging
from tqdm import tqdm
import xarray as xr

# sys.path = ["."] + sys.path
# from scripts import utils, load

"""
Warnings
"""
# warnings.simplefilter("ignore", UserWarning)


"""
Config
"""
# CONFIG = mngs.gen.load_configs()


"""
Functions & Classes
"""


def main():
    vizualize_colors(CC)
    pass


def vizualize_colors(colors):

    def gen_rand_sample(size=100):
        x = np.linspace(-1, 1, size)
        y = np.random.normal(size=size)
        s = np.random.randn(size)
        return x, y, s

    fig, axes = mngs.plt.subplots(ncols=4)

    for ii, (color_str, rgba) in enumerate(colors.items()):
        xx, yy, ss = gen_rand_sample()

        # Box color plot
        axes[0].rectangle(xx=ii, yy=0, ww=1, hh=1, color=rgba, label=color_str)

        # Line plot
        axes[1].plot_with_ci(xx, yy, ss, color=rgba, label=color_str)

        # Scatter plot
        axes[2].scatter(xx, yy, color=rgba, label=color_str)

        # KDE plot
        axes[3].kde(yy, color=rgba, label=color_str)

    for ax in axes.flat:
        # ax.axis("off")
        ax.legend()

    plt.tight_layout()
    plt.show()


if __name__ == "__main__":
    # # Argument Parser
    # import argparse
    # parser = argparse.ArgumentParser(description='')
    # parser.add_argument('--var', '-v', type=int, default=1, help='')
    # parser.add_argument('--flag', '-f', action='store_true', default=False, help='')
    # args = parser.parse_args()

    # Main
    CONFIG, sys.stdout, sys.stderr, plt, CC = mngs.gen.start(
        sys, plt, verbose=False
    )
    main()
    mngs.gen.close(CONFIG, verbose=False, notify=False)

# EOF



#!./env/bin/python3
# -*- coding: utf-8 -*-
# Time-stamp: "2024-10-13 19:04:32 (ywatanabe)"
# /home/ywatanabe/proj/mngs_repo/src/mngs/web/__init__.py

try:
    from ._summarize_url import summarize_url
except ImportError as e:
    pass
    # print(f"Warning: Failed to import summarize_url from ._summarize_url.")

# from ._summarize_url import summarize_url



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Time-stamp: "2024-09-13 19:16:42 (ywatanabe)"
# /home/ywatanabe/proj/mngs/src/mngs/ml/_gen_AI/_OpenAI.py

"""Imports"""

from mngs.ai._gen_ai._BaseGenAI import BaseGenAI
from openai import OpenAI as _OpenAI

"""Functions & Classes"""


class OpenAI(BaseGenAI):
    def __init__(
        self,
        system_setting="",
        model="",
        api_key="",
        stream=False,
        seed=None,
        n_keep=1,
        temperature=1.0,
        chat_history=None,
        max_tokens=4096,
    ):
        super().__init__(
            system_setting=system_setting,
            model=model,
            api_key=api_key,
            stream=stream,
            n_keep=n_keep,
            temperature=temperature,
            provider="OpenAI",
            chat_history=chat_history,
            max_tokens=max_tokens,
        )

    def _init_client(
        self,
    ):
        client = _OpenAI(api_key=self.api_key)
        return client

    def _api_call_static(self):
        kwargs = dict(
            model=self.model,
            messages=self.history,
            seed=self.seed,
            stream=False,
            temperature=self.temperature,
            max_tokens=self.max_tokens,
        )

        if kwargs.get("model") in ["o1-mini", "o1-preview"]:
            kwargs.pop("max_tokens")

        output = self.client.chat.completions.create(**kwargs)
        self.input_tokens += output.usage.prompt_tokens
        self.output_tokens += output.usage.completion_tokens

        out_text = output.choices[0].message.content

        return out_text

    def _api_call_stream(self):
        kwargs = dict(
            model=self.model,
            messages=self.history,
            max_tokens=4096,
            n=1,
            stream=self.stream,
            seed=self.seed,
            temperature=self.temperature,
            stream_options={"include_usage": True},
        )

        if kwargs.get("model") in ["o1-mini", "o1-preview"]:
            full_response = self._api_call_static()
            for char in full_response:
                yield char
            return

        stream = self.client.chat.completions.create(**kwargs)

        for chunk in stream:
            if chunk:
                try:
                    self.input_tokens += chunk.usage.prompt_tokens
                except:
                    pass
                try:
                    self.output_tokens += chunk.usage.completion_tokens
                except:
                    pass

                try:
                    current_text = chunk.choices[0].delta.content
                    if current_text:
                        yield f"{current_text}"
                except Exception as e:
                    pass


def main():
    model = "o1-mini"
    # model = "o1-preview"
    # model = "gpt-4o"
    stream = True
    max_tokens = 4906
    m = mngs.ai.GenAI(model, stream=stream, max_tokens=max_tokens)
    m("hi")


if __name__ == "__main__":
    import sys

    import matplotlib.pyplot as plt
    import mngs

    CONFIG, sys.stdout, sys.stderr, plt, CC = mngs.gen.start(
        sys, plt, verbose=False
    )
    main()
    mngs.gen.close(CONFIG, verbose=False, notify=False)

# EOF



#!./env/bin/python3
# -*- coding: utf-8 -*-
# Time-stamp: "2024-07-13 08:14:19 (ywatanabe)"
# Author: Yusuke Watanabe (ywata1989@gmail.com)

"""
This script does XYZ.
"""

# Imports
import matplotlib.pyplot as plt

from ._format_label import format_label


# Functions
def set_xyt(ax, x=False, y=False, t=False, format_labels=True):
    """Sets xlabel, ylabel and title"""

    if x is not False:
        x = format_label(x) if format_labels else x
        ax.set_xlabel(x)

    if y is not False:
        y = format_label(y) if format_labels else y
        ax.set_ylabel(y)

    if t is not False:
        t = format_label(t) if format_labels else t
        ax.set_title(t)

    return ax


if __name__ == "__main__":
    # Start
    CONFIG, sys.stdout, sys.stderr, plt, CC = mngs.gen.start(sys, plt)

    # (YOUR AWESOME CODE)

    # Close
    mngs.gen.close(CONFIG)

# EOF

"""
/ssh:ywatanabe@444:/home/ywatanabe/proj/entrance/mngs/plt/ax/_set_lt.py
"""



#!/./env/bin/python3
# -*- coding: utf-8 -*-
# Time-stamp: "2024-09-28 11:17:22 (ywatanabe)"
# /home/ywatanabe/proj/mngs_repo/src/mngs/pd/_to_xyz.py

import mngs
import numpy as np
import pandas as pd

def to_xyz(data_frame):
    """
    Convert a heatmap DataFrame into x, y, z format.

    Ensure the index and columns are the same, and if either exists, replace with that.

    Example
    -------
    data_frame = pd.DataFrame(...)  # Your DataFrame here
    out = to_xyz(data_frame)
    print(out)

    Parameters
    ----------
    data_frame : pandas.DataFrame
        The input DataFrame to be converted.

    Returns
    -------
    pandas.DataFrame
        A DataFrame formatted with columns ['x', 'y', 'z']
    """
    assert data_frame.shape[0] == data_frame.shape[1]

    if not data_frame.index.equals(data_frame.columns):
        if np.array_equal(data_frame.index, np.arange(len(data_frame.index))):
            data_frame.columns = data_frame.index
        elif np.array_equal(data_frame.columns, np.arange(len(data_frame.columns))):
            data_frame.index = data_frame.columns
        else:
            raise ValueError("Either index or columns must be a range of integers")

    x_name = data_frame.index.name or 'x'
    y_name = data_frame.columns.name or 'y'

    formatted_data_frames = []

    for column in data_frame.columns:
        column_data_frame = data_frame[column]
        formatted_data = pd.DataFrame({
            x_name: column_data_frame.index,
            y_name: column,
            'z': column_data_frame.values
        })
        formatted_data_frames.append(formatted_data)

    result = pd.concat(formatted_data_frames, ignore_index=True)
    result = mngs.pd.mv(result, 'z', -1)

    return result

# def to_xyz(data_frame):
#     """
#     Convert a heatmap DataFrame into x, y, z format.

#     Ensure the index and columns are the same, and if either exists, replace with that.

#     Example
#     -------
#     data_frame = pd.DataFrame(...)  # Your DataFrame here
#     out = to_xy(data_frame)
#     print(out)

#     Parameters
#     ----------
#     data_frame : pandas.DataFrame
#         The input DataFrame to be converted.

#     Returns
#     -------
#     pandas.DataFrame
#         A DataFrame formatted with columns ['x', 'y', 'z']
#     """
#     assert data_frame.shape[0] == data_frame.shape[1]

#     if not data_frame.index.equals(data_frame.columns):

#         if (data_frame.index == np.array(range(len(data_frame.index)))).all():
#             data_frame.columns = data_frame.index
#         elif (
#             data_frame.columns == np.array(range(len(data_frame.columns)))
#         ).all():
#             data_frame.index = data_frame.columns
#         else:
#             raise ValueError("Either index or columns must be a range of integers")

#     formatted_data_frames = []

#     for column in data_frame.columns:
#         column_data_frame = data_frame[column]
#         y_label = column_data_frame.name
#         column_data_frame = pd.DataFrame(column_data_frame)
#         column_data_frame["x"] = column_data_frame.index
#         column_data_frame["y"] = y_label
#         column_data_frame = column_data_frame.reset_index().drop(
#             columns=["index"]
#         )
#         column_data_frame = column_data_frame.rename(columns={y_label: "z"})
#         column_data_frame = mngs.pd.mv(column_data_frame, "z", -1)
#         formatted_data_frames.append(column_data_frame)

#     return pd.concat(formatted_data_frames, ignore_index=True)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Time-stamp: "2024-02-03 15:34:08 (ywatanabe)"

import matplotlib.pyplot as plt


def panel(tgt_width_mm=40, tgt_height_mm=None):
    """Creates a fixed-size ax figure for panels."""

    H_TO_W_RATIO = 0.7
    MM_TO_INCH_FACTOR = 1 / 25.4

    if tgt_height_mm is None:
        tgt_height_mm = H_TO_W_RATIO * tgt_width_mm

    # Convert target dimensions from millimeters to inches
    tgt_width_in = tgt_width_mm * MM_TO_INCH_FACTOR
    tgt_height_in = tgt_height_mm * MM_TO_INCH_FACTOR

    # Create a figure with the specified dimensions
    fig = plt.figure(figsize=(tgt_width_in * 2, tgt_height_in * 2))

    # Calculate the position and size of the axes in figure units (0 to 1)
    left = (fig.get_figwidth() - tgt_width_in) / 2 / fig.get_figwidth()
    bottom = (fig.get_figheight() - tgt_height_in) / 2 / fig.get_figheight()
    ax = fig.add_axes(
        [
            left,
            bottom,
            tgt_width_in / fig.get_figwidth(),
            tgt_height_in / fig.get_figheight(),
        ]
    )

    return fig, ax


if __name__ == "__main__":
    # Example usage:
    fig, ax = panel(tgt_width_mm=40, tgt_height_mm=40 * 0.7)
    ax.plot([1, 2, 3], [4, 5, 6])
    ax.scatter([1, 2, 3], [4, 5, 6])
    # ... compatible with other ax plotting methods as well
    plt.show()



import numpy as _np
import pandas as _pd
import xarray as _xr
import torch as _torch
from typing import List, Tuple, Dict, Any, Union, Sequence, Literal, Optional, Generator
from collections.abc import Iterable

ArrayLike = Union[
    List,
    Tuple,
    _np.ndarray,
    _pd.Series,
    _pd.DataFrame,
    _xr.DataArray,
    _torch.Tensor,
]


# from mngs.typing import List, Tuple, Dict, Any, Union, Sequence, Literal, Iterable, Generator, ArrayLike



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Time-stamp: "2024-03-05 13:17:04 (ywatanabe)"

# import warnings

import numpy as np
import pandas as pd
import torch


def to_sktime_df(X):
    """
    Converts a dataset to a format compatible with sktime, encapsulating each sample as a pandas DataFrame.

    Arguments:
    - X (numpy.ndarray or torch.Tensor or pandas.DataFrame): The input dataset with shape (n_samples, n_chs, seq_len).
      It should be a 3D array-like structure containing the time series data.

    Return:
    - sktime_df (pandas.DataFrame): A DataFrame where each element is a pandas Series representing a univariate time series.

    Data Types and Shapes:
    - If X is a numpy.ndarray, it should have the shape (n_samples, n_chs, seq_len).
    - If X is a torch.Tensor, it should have the shape (n_samples, n_chs, seq_len) and will be converted to a numpy array.
    - If X is a pandas.DataFrame, it is assumed to already be in the correct format and will be returned as is.

    References:
    - sktime: https://github.com/alan-turing-institute/sktime

    Examples:
    --------
    >>> X_np = np.random.rand(64, 160, 1024)
    >>> sktime_df = to_sktime_df(X_np)
    >>> type(sktime_df)
    <class 'pandas.core.frame.DataFrame'>
    """
    if isinstance(X, pd.DataFrame):
        return X
    elif torch.is_tensor(X):
        X = X.numpy()
    elif not isinstance(X, np.ndarray):
        raise ValueError(
            "Input X must be a numpy.ndarray, torch.Tensor, or pandas.DataFrame"
        )

    X = X.astype(np.float64)

    def _format_a_sample_for_sktime(x):
        """
        Formats a single sample for sktime compatibility.

        Arguments:
        - x (numpy.ndarray): A 2D array with shape (n_chs, seq_len) representing a single sample.

        Return:
        - dims (pandas.Series): A Series where each element is a pandas Series representing a univariate time series.
        """
        return pd.Series(
            [pd.Series(x[d], name=f"dim_{d}") for d in range(x.shape[0])]
        )

    sktime_df = pd.DataFrame(
        [_format_a_sample_for_sktime(X[i]) for i in range(X.shape[0])]
    )
    return sktime_df


# # Obsolete warning for future compatibility
# def to_sktime(*args, **kwargs):
#     warnings.warn(
#         "to_sktime is deprecated; use to_sktime_df instead.", FutureWarning
#     )
#     return to_sktime_df(*args, **kwargs)


# import pandas as pd
# import numpy as np
# import torch

# def to_sktime(X):
#     """
#     X.shape: (n_samples, n_chs, seq_len)
#     """

#     def _format_a_sample_for_sktime(x):
#         """
#         x.shape: (n_chs, seq_len)
#         """
#         dims = pd.Series(
#             [pd.Series(x[d], name=f"dim_{d}") for d in range(len(x))],
#             index=[f"dim_{i}" for i in np.arange(len(x))],
#         )
#         return dims

#     if torch.is_tensor(X):
#         X = X.numpy()
#         X = X.astype(np.float64)

#     return pd.DataFrame(
#         [_format_a_sample_for_sktime(X[i]) for i in range(len(X))]
#     )



#!./env/bin/python3
# -*- coding: utf-8 -*-
# Time-stamp: "2024-10-22 23:19:07 (ywatanabe)"
# /home/ywatanabe/proj/mngs_repo/src/mngs/io/_glob.py

import re
from natsort import natsorted
from glob import glob as _glob


def glob(expression, ensure_one=False):
    """
    Perform a glob operation with natural sorting and extended pattern support.

    This function extends the standard glob functionality by adding natural sorting
    and support for curly brace expansion in the glob pattern.

    Parameters:
    -----------
    expression : str
        The glob pattern to match against file paths. Supports standard glob syntax
        and curly brace expansion (e.g., 'dir/{a,b}/*.txt').

    Returns:
    --------
    list
        A naturally sorted list of file paths that match the given expression.

    Examples:
    ---------
    >>> glob('data/*.txt')
    ['data/file1.txt', 'data/file2.txt', 'data/file10.txt']

    >>> glob('data/{a,b}/*.txt')
    ['data/a/file1.txt', 'data/a/file2.txt', 'data/b/file1.txt']
    """
    glob_pattern = re.sub(r"{[^}]*}", "*", expression)
    try:
        found_paths = natsorted(_glob(eval(glob_pattern)))
    except:
        found_paths = natsorted(_glob(glob_pattern))

    if ensure_one:
        assert len(found_paths) == 1

    return found_paths

# import subprocess
# import re
# from natsort import natsorted

# def glob(expression, ensure_one=False):
#     """
#     Perform a file search operation using 'find' command with natural sorting and extended pattern support.

#     Parameters:
#     -----------
#     expression : str
#         The search pattern to match against file paths. Supports 'find' command syntax.

#     Returns:
#     --------
#     list
#         A naturally sorted list of file paths that match the given expression.

#     Examples:
#     ---------
#     >>> glob('data/*.txt')
#     ['data/file1.txt', 'data/file2.txt', 'data/file10.txt']

#     >>> glob('data/{a,b}/*.txt')
#     ['data/a/file1.txt', 'data/a/file2.txt', 'data/b/file1.txt']
#     """
#     glob_pattern = re.sub(r"{[^}]*}", "*", expression)
#     glob_pattern = eval(glob_pattern)
#     find_pattern = glob_pattern

#     cmd = f"find -wholename '*{find_pattern}' -type f -print0 | sort -z | xargs -0 ls -1d 2>/dev/null"
#     try:
#         result = subprocess.run(cmd, shell=True, check=True, stdout=subprocess.PIPE, universal_newlines=True)
#         found_paths = result.stdout.strip().split('\n')
#         found_paths = [path[2:] for path in found_paths if path]  # Remove './' prefix
#     except subprocess.CalledProcessError:
#         found_paths = []

#     found_paths = natsorted(found_paths)

#     if ensure_one:
#         assert len(found_paths) == 1

#     return found_paths



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Time-stamp: "2024-03-30 07:26:35 (ywatanabe)"

import torch.nn as nn


class TransposeLayer(nn.Module):
    def __init__(
        self,
        axis1,
        axis2,
    ):
        super().__init__()
        self.axis1 = axis1
        self.axis2 = axis2

    def forward(self, x):
        return x.transpose(self.axis1, self.axis2)



#!./env/bin/python3
# -*- coding: utf-8 -*-
# Time-stamp: "2024-09-28 10:19:09 (ywatanabe)"
# Author: Yusuke Watanabe (ywata1989@gmail.com)

"""This script does XYZ."""

import sys

import matplotlib.pyplot as plt
import mngs
import numpy as np


# Functions
def imshow2d(
    ax,
    arr_2d,
    cbar=True,
    cbar_label=None,
    cbar_shrink=1.0,
    cbar_fraction=0.046,
    cbar_pad=0.04,
    cmap="viridis",
    aspect="auto",
    vmin=None,
    vmax=None,
    **kwargs,
):
    """
    Imshows an two-dimensional array with theese two conditions:
    1) The first dimension represents the x dim, from left to right.
    2) The second dimension represents the y dim, from bottom to top
    """

    assert arr_2d.ndim == 2

    # Transposes arr_2d for correct orientation
    arr_2d = arr_2d.T

    # Cals the original ax.imshow() method on the transposed array
    im = ax.imshow(
        arr_2d, cmap=cmap, vmin=vmin, vmax=vmax, aspect=aspect, **kwargs
    )

    # Color bar
    if cbar:
        fig = ax.get_figure()
        _cbar = fig.colorbar(
            im, ax=ax, shrink=cbar_shrink, fraction=cbar_fraction, pad=cbar_pad
        )
        if cbar_label:
            _cbar.set_label(cbar_label)

    # Invert y-axis to match typical image orientation
    ax.invert_yaxis()

    return ax


if __name__ == "__main__":
    # Start
    CONFIG, sys.stdout, sys.stderr, plt, CC = mngs.gen.start(sys, plt)

    # Example usage
    fig, ax = plt.subplots()
    data = np.random.rand(10, 20)  # Random data
    imshow2d(ax, data)
    plt.show()

    # Close
    mngs.gen.close(CONFIG)

# EOF

"""
/ssh:ywatanabe@444:/home/ywatanabe/proj/entrance/mngs/plt/ax/_imshow2d.py
"""
